// Compression.cpp: implementation of the CCompression class.
//
// This class compress the images that were acquired from the frame grabber
// using any VCM Codec you specify through the fourcc_codec parameter in the
// Init() function. Please note that temporal compression codecs do not work,
// just keyframe codecs like e.g. HuffYUV and MJPG are supported.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdio.h>
#include <stdarg.h>
#include <vfw.h> // platform dependent: need video for windows header file
#include "Grabsrv.h"
#include "headers.h"
#include "Compression.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////
// 
// In the construction of the class just some variables are initialized.
//
CCompression::CCompression()
{
 // set handle to codec instance to NULL
 m_pCompressor = NULL;
 
 // set pointers to formats to NULL
 m_pInputFmt = NULL;
 m_pOutputFmt = NULL;
 
 // we are not yet initialized
 m_bInitialized = FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// In the destructor, we check whether the user has forgot to call Done() before the
// destructor is called. If the user has not called Done, we do all necessary cleanup
// here.
//
CCompression::~CCompression()
{
 if (m_bInitialized == TRUE) // we are still initialized ? 
 {
  DWORD ret = ICCompressEnd( m_pCompressor ); // call ICCompressEnd on codec instance
  if (ret != ICERR_OK) myErrorHandler(-1, "~CCompression: Error %d (%xh) on ICCompressEnd !", ret, ret);
  ret = ICClose( m_pCompressor ); // close the codec instance
  if (ret != ICERR_OK) myErrorHandler(-1, "~CCompression: Error %d (%xh) on ICClose !", ret, ret);
  m_pCompressor = NULL; // set to NULL (uninitalized)
 }

 // free input and output formats if they are initialized
 if (m_pInputFmt != NULL) memman.sw_free( (void *) m_pInputFmt );
 if (m_pOutputFmt != NULL) memman.sw_free( (void *) m_pOutputFmt );
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// In this function, the codec is initialized and it is checked whether the codec accepts
// the inputFmt image format. The function returns TRUE if everything went fine, FALSE 
// otherwise.
//
// Parameters:
//  fourcc_codec  - four character code of the codec to be instantiated
//  inputFmt      - pointer to BITMAPINFOHEADER describing the frame (image) format
//
// Return value:
//  TRUE          - everything went fine, codec is instantiated and bound to the input format
//  FALSE         - there was an error instantiating the codec or the input format was not
//                  valid
//  More specific error codes are written out through the myErrorHandler function.
//
// Remark:
//  At the moment only keyframe codecs are supported. Temporal compression (delta-frame) 
//  codecs do not work !
//
BOOL CCompression::Init( DWORD fourcc_codec, LPBITMAPINFOHEADER inputFmt)
{
    DWORD ret;

    if (m_bInitialized == FALSE) // we are not initialized, are we ? 
	{

      if (inputFmt == NULL) // if there is no input format specified, return error
	  {
       myErrorHandler(-1, "CCompression::Init: Parameter inputFmt is NULL !");
       
	   // return error
       return FALSE;
	  }

      // allocate memory needed for input format
	  m_pInputFmt = (LPBITMAPINFOHEADER) memman.sw_malloc( inputFmt->biSize );
	  if (m_pInputFmt == NULL) // could not allocate -> print error
	  {
       myErrorHandler(-1, "CCompression::Init: Could not allocate memory needed for frame input format.");
       
	   // return error
       return FALSE;
	  }
	
	  // copy the input format to the internal memory location
	  memcpy( (void *) m_pInputFmt, (const void *) inputFmt, inputFmt->biSize );

      // open (instantiate) the codec
	  m_pCompressor = ICOpen(ICTYPE_VIDEO, fourcc_codec, ICMODE_FASTCOMPRESS);

	  if (m_pCompressor == NULL) // could not open ? 
	  {
       // print error message
	   myErrorHandler(-1, "CCompression::Init: Could not open the Codec '%c%c%c%c'.", (fourcc_codec)&0xFF, (fourcc_codec>>8)&0xFF,(fourcc_codec>>16)&0xFF,(fourcc_codec>>24)&0xFF);
	   // free previously allocated input format
	   memman.sw_free((void *)m_pInputFmt);
	   m_pInputFmt = NULL;
       
	   // return error
       return FALSE;
	  }

      m_iFourccCodec = fourcc_codec;

      // get the length of the output format  
      int lFmtLength = ICCompressGetFormat(m_pCompressor, m_pInputFmt, NULL);

      // allocate internal storage location for the output format
	  m_pOutputFmt = (LPBITMAPINFOHEADER) memman.sw_malloc( lFmtLength );
	  if (m_pOutputFmt == NULL) // could not allocate ? 
	  {
       // print error 
	   myErrorHandler(-1, "CCompression::Init: Could not allocate memory needed for frame output format.");
       
	   // do cleanup
	   ICClose(m_pCompressor);
	   m_pCompressor = NULL;
	   memman.sw_free((void *)m_pInputFmt);
	   m_pInputFmt = NULL;
       
	   // return error
       return FALSE;
	  }

	  m_pOutputFmt->biSize = lFmtLength;

      // get output format into the previously allocated memory location 
      ret = ICCompressGetFormat(m_pCompressor, m_pInputFmt, m_pOutputFmt);
	  if (ret != ICERR_OK) // error happened ? 
	  {
       // print error
	   myErrorHandler(-1, "CCompression::Init : Could not get frame output format (errcode=%xh).", ret);        
       
	   // do cleanup
	   ICClose(m_pCompressor);
       m_pCompressor = NULL;
	   memman.sw_free((void *)m_pInputFmt);
	   m_pInputFmt = NULL;
	   memman.sw_free((void *)m_pOutputFmt);
	   m_pOutputFmt = NULL;
       
	   // return error
       return FALSE;
	  }

      // start compression; after successfully calling this, we can 
	  // compress the frames
      ret = ICCompressBegin(m_pCompressor, m_pInputFmt, m_pOutputFmt);
	  if (ret != ICERR_OK) // error ? 
	  {
       // print error
	   myErrorHandler(-1, "CCompression::Init : Error on ICCompressBegin (errcode=%xh).", ret);        
       
	   // do cleanup
	   ICClose(m_pCompressor);
       m_pCompressor = NULL;
	   memman.sw_free((void *)m_pInputFmt);
	   m_pInputFmt = NULL;
	   memman.sw_free((void *)m_pOutputFmt);
	   m_pOutputFmt = NULL;
       
	   // return error
	   return FALSE;
	  }

      // everything went fine, we are now initialized
	  m_bInitialized = TRUE;
	}

	return TRUE; // success !
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// In this function, the frame deliverd in the memory location 'frame' is compressed.
// The output will be written to the memory location 'compressedframe' points to.
// You have to make sure that the memory location for the compressed frame is large
// enough to hold the maximum compressed frame size. You can obtain the maximum size
// of a compressed frame by querying the GetMaxCompressedSize() function.
//
// Parameters:
//  frame            - read: uncompressed frame data (bits)
//  compressedframe  - write: compressed frame data (bits)
//
// Return value:
//  TRUE   : everything went fine
//  FALSE  : there was an error while compressing
//
BOOL CCompression::CompressFrame( char *frame, char *compressedframe )
{
 if (m_bInitialized == TRUE) // are we initialized
 {
     DWORD retck;
     DWORD retflags;
	 
	 // compress the frame using windows compression manager api call
	 DWORD ret = ICCompress( m_pCompressor, ICCOMPRESS_KEYFRAME, m_pOutputFmt, (LPVOID) compressedframe, m_pInputFmt, (LPVOID) frame, &retck, &retflags, 1, 0, 100, NULL, NULL);
	 if (ret == ICERR_OK) // everything went fine ? 
	 {
	  return TRUE; // return success
	 }
 }

 return FALSE; // return error
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// Using this function, you can gain the information how big the output format 
// structure can get. You can use this information to allocate a buffer that is large 
// enough to store the output format.
// 
// Parameter:
//   length   :  write-back variable to store the size of the output format
// 
// Return value:
//   TRUE     :  everything went fine, proper value is delivered in '*length'
//   FALSE    :  error while getting output format size (the output format is not 
//               initialized yet)
//
BOOL CCompression::GetOutputFormatLength( int *length )
{
 if (m_pOutputFmt != NULL)
 {
  *length = m_pOutputFmt->biSize;
  return TRUE;
 }
  
 myErrorHandler(-1, "CCompression::GetOutputFormatLength: Error: output format not initialized.");
 *length = 0;
 return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// Using this function, you can obtain the current output format of the codec. A 
// pointer to a memory location at least as big as the length GetOutputFormatLength()
// returns to you has to be passed to this function. You also have to fill in the 
// biSize-property of your BITMAPINFOHEADER-structure (outputFmt parameter). The size 
// property must match the size of the output format (the size you have obtained by 
// calling GetOutputFormatLength().
// 
// Parameter:
//   outputFmt :  read/write - pointer to memory location where the output format 
//                             will be stored
// 
// Return value:
//   TRUE   :  everything went fine, proper output format is delivered in 'outputFmt'
//   FALSE  :  error while getting output format (e.g. you have passed NULL or you have 
//             put the wrong size in the biSize property of the outputFmt structure)
//
BOOL CCompression::GetOutputFormat( LPBITMAPINFOHEADER outputFmt )
{
 if (m_pOutputFmt != NULL) // valid output format pointer ? 
 {
  if (m_pOutputFmt->biSize != outputFmt->biSize) // does the size not match ? 
  {
	  myErrorHandler(-1, "CCompression::GetOutputFormat: Size of the output format does not match");
	  return FALSE; // failure
  }
 
  // everything fine, copy the outputFmt structure
  memcpy((void *) outputFmt, (const void *)m_pOutputFmt, m_pOutputFmt->biSize);
  
  return TRUE; // success
 }
  
 myErrorHandler(-1, "CCompression::GetOutputFormat: Error: output format not initialized.");
 return FALSE; // failure
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// This function should be called after you have finished compressing the frames. It
// does all necessary cleanup (memory deallocation) and closes the codec instance.
// 
// Return value:
//   TRUE   :  everything went fine, codec is closed and class in 'uninitialized' state
//   FALSE  :  error while cleaning up, class is _not_ uninitialized (due to my experience
//             this should not happen)
//
BOOL CCompression::Done()
{
 DWORD ret;
 BOOL retval=TRUE;

 if (m_bInitialized == TRUE) // are we initialized ? 
 {
   ret = ICCompressEnd(m_pCompressor); // call 'end-of-compression-session'
   if (ret != ICERR_OK) // error ? 
   {
	   myErrorHandler(-1, "CCompression::Done : Error on ICCompressEnd (errorcode=%xh)", ret);
	   retval = FALSE;
   }

   ret = ICClose(m_pCompressor); // close the codec
   if (ret != ICERR_OK) // error ? 
   {
       myErrorHandler(-1, "CCompression::Done : Error calling ICClose (errorcode=%xh)", ret);
	   retval = FALSE;
   }

   // input format allocated ? -> free it 
   if (m_pInputFmt != NULL) 
	    memman.sw_free( (void *) m_pInputFmt ); 
   m_pInputFmt = NULL;

   // output format allocated ? -> free it 
   if (m_pOutputFmt != NULL) 
	    memman.sw_free( (void *) m_pOutputFmt ); 
   m_pOutputFmt = NULL;

   // everything went fine: set class to 'uninitialized' state
   if (retval == TRUE) m_bInitialized = FALSE;
 }
 
 // return success or failure
 return retval;
}
/////////////////////////////////////////////////////////////////////////////////////
// 
// This function is needed to output errors that happen while executing the program.
// A error message is constructed out of the parameters and displayed using a message 
// box.
// 
// Parameter:
//   errornum    :  additional error number, is not really used at the moment
//   format, ... :  same like printf()
// 
// Return value:
//   always 1
//
// Remarks:
//   Please not that the string you construct should not be larger than 511 characters
//   (512 including the '0' at the end). Otherwise the string will be truncated. 
//
int CCompression::myErrorHandler( int errornum, const char *format, ... )
{
 char m_msg[512];

 // print the arguments	into a single string
 va_list args;
 va_start(args, format);
 _vsnprintf(m_msg, 511, format, args);
 va_end(args);
 
 // set last error (not used at the moment)
 m_iLastError = errornum;

 // open message box 
 AfxMessageBox( m_msg, MB_ICONEXCLAMATION | MB_OK );

 return 1;
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// Using this function, you can gain the information how big a compressed frame can 
// get. You can use this information to allocate a buffer that is large enough to 
// store a compressed frame.
// 
// Parameter:
//   size     :  write-back variable to store the maximum size of a compressed frame
// 
// Return value:
//   TRUE     :  everything went fine, proper value is delivered in '*size'
//   FALSE    :  error while getting maximum compressed size (the codec is not 
//               initialized yet)
//
BOOL CCompression::GetMaxCompressedSize( int *size )
{
 if (m_bInitialized == TRUE) // are we initialized ? 
 {
	// get the size
	DWORD ret = ICCompressGetSize( m_pCompressor, m_pInputFmt, m_pOutputFmt);
    // return if pointer != NULL
	if (size != NULL) *size = ret;

    return TRUE; //success
 }

 return FALSE; // failue
}

/////////////////////////////////////////////////////////////////////////////////////
// ReturnFourccCodec()
// 
// Using this function, you can obtain the four character code associated with the 
// codec used to compress the video data.
// 
// If the class was not initialized (using Init() method), the function returns 0, 
// otherwise the four character code of the codec is returned.
//
DWORD CCompression::ReturnFourccCodec()
{
 if (m_bInitialized == TRUE) // successfully initialized ? 
 {
  return( m_iFourccCodec ); // return fourcc 
 }
 
 return( 0 ); // not initialized? -> return 0
}
