// Grab.cpp: implementation of the CGrab class.
//
// based on code from BESSY framegrabbing software "Pitzcam". 
// written by Stefan Weisse (sweisse@ifh.de)
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "headers.h"
#include <vfw.h>
#include "Queue.h"
#include "Grabsrv.h"
#include "Grab.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
//
// This class is meant as an encapsulation of the raw ITEX grabber 
// function which are needed to access the PCVision frame grabber 
// card. Using this class one is able to snap pictures, grab frames
// continuously and switch the camera port on the frame grabber card.
// The class issues a PostMessage (WM_COMMAND, ID_NOTIFY_GRABBED) to
// the main application dialog so that the main class is notified 
// whether there is a new frame grabbed. Then the main class can call
// the getActImageData() function to get all important data from the
// grabbed frame.
//
// The lifecycle is as follows:
// - instantiate the class
// - call Init() to initialize the class and the ITEX system
// - call StartGrab/StopGrab to start/stop continuous grabbing or
// - call Snap() to acquire a single image
// - call setCamera / getCamera to set or retrieve the camera port
// - when everything is finished, call Done() to deinitialize the 
//   class
//
//////////////////////////////////////////////////////////////////////


// define this variable if the grabbing uses a trigger
// undefine this variable if the grabbing uses no trigger (ie. automatic write enable)
#define _TRIGGER_
//#undef _TRIGGER_

// define if you want ITEX and Grabbing error output ( debug )
// undefine if you do not want error output (release version)
//#define _SW_DEBUG_
#undef _SW_DEBUG_

#define _REAL_RATE_ 5 // 10 Hz

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// 
// In construction of the class, just some variables are initialized to
// their default values.
CGrab::CGrab()
{
 // we are not grabbing at the moment
 m_bIsGrabbing = FALSE;
 
 // no grab thread is up and running
 m_pGrabThread = NULL;
 
 m_iLastError = 0; // no error
 
 // initialize width and height values
 m_iFrameWidth = 0;
 m_iFrameHeight=0;

 // set camera port to 0
 m_iCamera = 0;

 // class is not yet initialized
 m_bSuccessfulInitialized = FALSE;
 
 // up to now no double buffer is allocated
 m_pDoubleImageBuffer = NULL;
 
 // starting frame number is zero
 m_iFrameNumber = 0;
}

//////////////////////////////////////////////////////////////////////
// 
// During destruction, we check whether the class is initialized and
// grabbing. If any of the conditions are true, we stop grabbing and/or
// call Done() to deinitialize the class.
//
CGrab::~CGrab()
{
 if (m_bSuccessfulInitialized == TRUE) // are we initialized?
 {
	if (m_bIsGrabbing == TRUE) // are we grabbing ? 
	{
     StopGrab(); // stop grabbing
    }

	Done(); // deinitialize class and ITEX
 }
}

//////////////////////////////////////////////////////////////////////
// 
// The function setCamera() is needed to set the camera port on the 
// PCVision frame grabber card. The parameter cam can have integer
// values of 0 to 3. The function returns TRUE if everything went 
// fine and false otherwise.
//
BOOL CGrab::setCamera( int cam )
{
	m_iCamera = cam;

    return TRUE; //success
}

//////////////////////////////////////////////////////////////////////
// 
// The function getCamera() retrieves the actually selected camera 
// port. The camera port is returned. If the function returns '-1'
// then something went wrong. Either the ITEX system is not initialized
// or the class was not able to retrieve the camera port.
//
int CGrab::getCamera()
{
	return( m_iCamera ); // return camera port
}

//////////////////////////////////////////////////////////////////////
// 
// The function Init() initializes the ITEX system and the class so 
// that we are able to snap/grab frames and are able to get and set
// the camera port value. The two parameters are informations passed
// to the underlying ITEX system. In normal case, both system and 
// boardno are zero. The function returns TRUE if the class and the
// ITEX system is successfully initialized; FALSE otherwise.
//
BOOL CGrab::Init( short system, short boardno, BOOL exttrig, BOOL trigmod5, CQueue *grab_queue, int nref0, int pref0 )
{
    int x;

	FILE *fp = fopen("spot_11_07_2002_150_pictures_take2.imm", "rb");

	m_videoframes = new CQueue();

	for (int i=0;i<150;i++)
	{
      char *frame = (char *) memman.sw_malloc(768*574);

	  fread( (void *) &x, 4, 1, fp);
	  fread( (void *) &x, 4, 1, fp);

      fread( (void *) frame, 768*574, 1, fp);

	  fread( (void *) &x, 4, 1, fp);
	  fread( (void *) &x, 4, 1, fp);

	  //memcpy((void *) frame, (const void *) (temp+8), 768*574);

	  m_videoframes->AddTail( (void *) frame );
	}

	fclose(fp);

    m_GrabQueue = grab_queue;

    m_iFrameWidth=768;
	m_iFrameHeight=574;

    // allocate memory needed for storing up to two frames (double buffering)
    int dbuflen = m_iFrameWidth*m_iFrameHeight*2;
	m_pDoubleImageBuffer = (unsigned char *) memman.sw_malloc( dbuflen ); // 2 images grayscale (8bit)
	if (m_pDoubleImageBuffer == NULL)
	{
	 myErrorHandler(17, "Out of memory allocating %d bytes for double image buffer (CGrab::Init)", dbuflen);
	 delete m_videoframes;
	 return FALSE;
	}
	
	// when we are here everything went fine
    m_bSuccessfulInitialized = TRUE;
 
    // success
    return TRUE;
}

//////////////////////////////////////////////////////////////////////
// 
// This function is needed for snapping a single image. After the 
// image is snapped successfully, the function posts a message to
// the main window that a new frame has been acquired.
//
// The function returns TRUE if everything went fine or FALSE other-
// wise.
//
BOOL CGrab::Snap()
{
	return(TRUE); // success
}

//////////////////////////////////////////////////////////////////////
// 
// This function starts the continuous grabbing of images. It checks 
// whether everything is properly initialized and starts up the grabber
// thread which acquires the images. The function returns TRUE if 
// grabbing is started, FALSE otherwise.
// 
BOOL CGrab::StartGrab()
{
 // check whether class is properly initialized
 if (m_bSuccessfulInitialized == FALSE) return FALSE;
 
 // if we are actually grabbing, do nothing
 if (m_bIsGrabbing == TRUE) return(TRUE);

 // start the grabbing
 m_hgrab = 1;
 
 // fire up grabber thread with time critical priority
 m_bFirstTime = TRUE;
 m_pGrabThread = AfxBeginThread(threadGrab, this, THREAD_PRIORITY_NORMAL); 
 if (!m_pGrabThread) // error ? 
 {
  myErrorHandler(11, "Could not create grabber thread.");
  return( FALSE ); // failure
 }
 
 // do not auto-delete grabber thread
 m_pGrabThread->m_bAutoDelete = false;
 
 // set grabbing flag to true
 m_bIsGrabbing = TRUE;

 return( TRUE ); // success
}

//////////////////////////////////////////////////////////////////////
// 
// This function stops the continuous grabbing of images. It checks 
// whether we are grabbing and stops the grabbing if necessary. The 
// function returns always TRUE.
//
BOOL CGrab::StopGrab()
{
 if (m_bIsGrabbing == TRUE) // are we grabbing
 {
  // set grabbing handle to zero (uninitialized), so that
  // the thread knows that it should stop
  m_hgrab = 0;
  // wait for the thread to come down
  WaitForSingleObject(m_pGrabThread->m_hThread, INFINITE);
  // delete grab thread
  delete m_pGrabThread;
  m_pGrabThread = 0;
  
  // we are not grabbing any more
  m_bIsGrabbing = FALSE;
 }

 return( TRUE ); // success
}

//////////////////////////////////////////////////////////////////////
// 
// This function deinitializes the class and the ITEX grabbing system.
// After a call to this function, the class is going into the unin-
// intialized state. You can not grab, snap, get/set camera port any
// more. The function always returns TRUE. The two sleeps are there 
// because without them, ITEX was not properly uninitialized sometimes.
// ITEX could be uninitialized but not initialized any more.
//
BOOL CGrab::Done()
{
  if (m_bSuccessfulInitialized == TRUE)
  {
   int length = m_videoframes->GetSize();

   for (int i=0;i<length;i++)
   {
	   void *ptr = m_videoframes->RemoveHead(); 
	   if (ptr != NULL) memman.sw_free(ptr);
   }

   delete m_videoframes;
   m_videoframes = NULL;

   // free memory needed by the double buffer
   memman.sw_free( (void *) m_pDoubleImageBuffer );
  
   // class is not successfully initialized any more
   m_bSuccessfulInitialized = FALSE;
  }
  
  return( TRUE ); //success
}

//////////////////////////////////////////////////////////////////////
// 
// This function returns whether the class is actually grabbing images.
// It returns TRUE if grabbing is up and running, FALSE otherwise.
//
BOOL CGrab::IsGrabbing()
{
 return( m_bIsGrabbing );
}

//////////////////////////////////////////////////////////////////////
// 
// This is the thread which is needed for coninuous acquisition of 
// frames. After having successful acquired one frame, a message is
// posted to the main window indicating that a new frame was grabbed.
// To stop the thread just set the member m_hgrab to zero.
//
UINT CGrab::threadGrab(LPVOID pParam)
{
    BYTE *ptr=NULL;
    double passedmilliseconds;
	CGrab *_this = (CGrab *)pParam;
    int tmp=0;
	unsigned int sleeprate = (int) (1000.0 / ((double)_REAL_RATE_))+1;

	while (_this->m_hgrab)
	{
        // timestamp and cameranumber
		// Wait for the next image to be ready and transfer to host
		ptr = (BYTE *) _this->m_videoframes->RemoveHead();
		if (ptr == NULL) continue;
		_this->m_videoframes->AddTail((void *) ptr);
		if (!_this->m_hgrab) return 0;

		 _this->m_pAktImage = (BYTE *) ptr;
		 //_this->m_iFrameNumber++;
		 ::_ftime( &_this->m_tTimeStamp );

   		  // Notify our main window that we grabbed a new image
		  // grabbed image can be obtained using getActImageData-Function
    	  _this->m_iFrameNumber++;

		  /* calculate framerate */
	      // calculate gap between frames (timestamp)

    	  if (_this->m_bFirstTime == TRUE)
		  {
            memcpy((void *) &_this->m_tVeryFirstTimestamp, (const void *) &_this->m_tTimeStamp, sizeof(_timeb));
	        _this->m_bFirstTime = FALSE;
			_this->m_iFirstFrameNo = _this->m_iFrameNumber;
		  }
		  else
		  {
			// difference between timestamps
			passedmilliseconds = difftime(_this->m_tTimeStamp.time, _this->m_tVeryFirstTimestamp.time) * 1000.0;

			passedmilliseconds -= ((double)_this->m_tVeryFirstTimestamp.millitm);
			passedmilliseconds += ((double)_this->m_tTimeStamp.millitm);
		  }

          _this->m_framespersec = 1000.0 / ( passedmilliseconds / ((double)(_this->m_iFrameNumber-_this->m_iFirstFrameNo+1)));
		  /* calculate framerate end */

          char *block = (char *) memman.sw_malloc( sizeof(VIDHDR)+(_this->m_iFrameWidth*_this->m_iFrameHeight*1) );

		  if (block != NULL)
		  {
			  VIDHDR hdr;
	          hdr.kennung = 0x11223344;
		      hdr.compression = MAKEFOURCC('H','F','Y','U');
		      hdr.cameraport = _this->m_iCamera;
		      hdr.frameno = _this->m_iFrameNumber;
		      hdr.len_data = _this->m_iFrameWidth*_this->m_iFrameHeight;
		      hdr.len_fmthdr = 0;
		      hdr.len_total = hdr.len_data + hdr.len_fmthdr;
		      hdr.widthframe = _this->m_iFrameWidth;
		      hdr.heightframe = _this->m_iFrameHeight;
		      hdr.framerate = _this->m_framespersec;
		      hdr.compressed = 0;
		      hdr.bpp = 8;
		      memman.memcpysw((char *) &hdr.timestamp, (char *) &_this->m_tTimeStamp, sizeof(_timeb) );  

              memman.memcpysw( block, (char *) &hdr, sizeof(VIDHDR) );
		      memman.memcpysw( block+sizeof(VIDHDR), (char *) _this->m_pAktImage, _this->m_iFrameWidth*_this->m_iFrameHeight );
   
              _this->m_GrabQueue->AddTail( (void *) block );
          
		      AfxGetMainWnd()->PostMessage(WM_COMMAND, ID_NOTIFY_GRABBED);
          }
		  ::Sleep(sleeprate);
		  //::Sleep(10000);

          //if (_this->m_framespersec > ((double)_REAL_RATE_)) sleeprate+=10; else sleeprate-=10;
		  //if (sleeprate <10) sleeprate = 10;

	}
	return 0;
}


//////////////////////////////////////////////////////////////////////
// 
// This function is called by the ITEX system when an error takes 
// place. If you have defined the _SW_DEBUG_ variable, the message
// will be printed in a message box. 
//
int CGrab::itexErrorHandler( const char *format, va_list marker)
{
#ifdef _SW_DEBUG_
 char s[200];

 // print error message
 int l = _vsnprintf(s, 198, format, marker);
 
 // open message box
 AfxMessageBox( s, MB_ICONEXCLAMATION | MB_OK );

 return l;
#else
 return( 0 );
#endif
}

//////////////////////////////////////////////////////////////////////
// 
// This function is the error handler for errors happening at class 
// level. The parameter errornum is not really used at the moment. The 
// other parameters ( format, ...) are just like printf().
//
int CGrab::myErrorHandler( int errornum, const char *format, ... )
{
#ifdef _SW_DEBUG_
 char m_msg[512];
	
 va_list args;
 va_start(args, format);
 _vsnprintf(m_msg, 198, format, args);
 va_end(args);
 
 m_iLastError = errornum;
 AfxMessageBox( m_msg, MB_ICONEXCLAMATION | MB_OK );
#endif 

 return 1;
}

//////////////////////////////////////////////////////////////////////
// 
// This function returns the length in bytes of an acquired frame. If 
// the class is not yet initialized, the function returns zero.
//
int CGrab::GetFrameLength()
{
 if (m_bSuccessfulInitialized == TRUE) // successfully initialized ? 
 {
  return( m_iFrameWidth * m_iFrameHeight ); // return frame length
 }
  else
	  return 0; // return zero
}

BOOL CGrab::getNRef( short &nref)
{
 nref = 234;
 return( TRUE );
}
  
BOOL CGrab::setNRef( short nref )
{
 return TRUE;
}

BOOL CGrab::getPRef( short &pref )
{
 pref = 782;
 return TRUE;
}

BOOL CGrab::setPRef( short pref )
{
 return TRUE;
}

int CGrab::getFrameWidth()
{
 return( m_iFrameWidth );
}

int CGrab::getFrameHeight()
{
 return( m_iFrameHeight );
}

int CGrab::getFrameBpp()
{
 return( 8 );
}
