// OrientationHelper.cpp: implementation of the COrientationHelper class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Grabsrv.h"
#include "OrientationHelper.h"
#include "Headers.h"
#include "MemoryManagement.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// COrientationHelper - rotates or mirrors a video frame
//

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COrientationHelper::COrientationHelper()
{
}

COrientationHelper::~COrientationHelper()
{
}

//////////////////////////////////////////////////////////////////////
// CalcOrientation - rotates or mirrors the video frame bits based on
//                   the information in string 'orientation'
//
// Parameters:
//   dst           - destination pointer (must be at least widthbefore*heightbefore bytes long)
//   src           - destination pointer (must be at least widthbefore*heightbefore bytes long)
//  orientation    - describes the type of calculation that is done on the source array
//                   "ROT90CW" - rotates source frame by 90 clockwise
//                   "ROT180CW" - rotates source frame by 180 
//                   "ROT270CW" - rotates source frame by 270 clockwise
//                   "MIRRORHORIZ" - mirrors the source frame along horizontal axis
//                   "MIRRORVERT" - mirrors the source frame along vertical axis
// widthbefore     - width in pixels of the source image before applying the transformation
// heightbefore    - height in pixels of the source image before applying the transformation
//
void COrientationHelper::CalcOrientation(char *dst, char *src, CString orientation, int widthbefore, int heightbefore, int physbpp)
{
	if (orientation == CString("ROT90CW")) COrientationHelper::ROT90CW( dst, src, widthbefore, heightbefore, physbpp );
	if (orientation == CString("ROT180CW")) COrientationHelper::ROT180CW( dst, src, widthbefore, heightbefore, physbpp );
	if (orientation == CString("ROT270CW")) COrientationHelper::ROT270CW( dst, src, widthbefore, heightbefore, physbpp );
	if (orientation == CString("MIRRORHORIZ")) COrientationHelper::MIRRORHORIZ( dst, src, widthbefore, heightbefore, physbpp );
	if (orientation == CString("MIRRORVERT")) COrientationHelper::MIRRORVERT( dst, src, widthbefore, heightbefore, physbpp );
	if (orientation == CString("ROT90CWMIRRHORIZ")) COrientationHelper::ROT90ANDMIRRHORIZ( dst, src, widthbefore, heightbefore, physbpp);
}

////////////////////////////////////////////////////////////////////////
// ROT90CW - rotates the image by 90 clockwise
//
// Parameters:
//  dst          - destination array
//  src          - source array
//  widthbefore  - width of the source image in pixels
//  heightbefore - height of the source image in pixels
//
void COrientationHelper::ROT90CW( char *dst, char *src, int widthbefore, int heightbefore, int physbpp )
{
	if (physbpp <= 8)
	{
		const unsigned char *source = (unsigned char *) src;
		unsigned char *dest = (unsigned char *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}
	else if (physbpp <= 16)
	{
		const unsigned short *source = (unsigned short *) src;
		unsigned short *dest = (unsigned short *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}
	else
	{
		const unsigned int *source = (unsigned int *) src;
		unsigned int *dest = (unsigned int *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}

}

////////////////////////////////////////////////////////////////////////
// ROT180CW - rotates the image by 180
//
// Parameters:
//  dst          - destination array
//  src          - source array
//  widthbefore  - width of the source image in pixels
//  heightbefore - height of the source image in pixels
//
void COrientationHelper::ROT180CW( char *dst, char *src, int widthbefore, int heightbefore, int physbpp )
{
	if (physbpp <= 8)
	{
		const unsigned char *source = (unsigned char *) src;
		source += widthbefore*heightbefore;
		unsigned char *dest = (unsigned char *) dst;

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}
	else if (physbpp <= 16)
	{
		const unsigned short *source = (unsigned short *) src;
		source += widthbefore*heightbefore;
		unsigned short *dest = (unsigned short *) dst;

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}
	else
	{
		const unsigned int *source = (unsigned int *) src;
		source += widthbefore*heightbefore;
		unsigned int *dest = (unsigned int *) dst;

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}

}

////////////////////////////////////////////////////////////////////////
// ROT270CW - rotates the image by 270 clockwise
//
// Parameters:
//  dst          - destination array
//  src          - source array
//  widthbefore  - width of the source image in pixels
//  heightbefore - height of the source image in pixels
//
void COrientationHelper::ROT270CW( char *dst, char *src, int widthbefore, int heightbefore, int physbpp )
{
	if (physbpp <= 8)
	{
		const unsigned char *source = (unsigned char *) src;
		unsigned char *dest = (unsigned char *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}
	else if (physbpp <= 16)
	{
		const unsigned short *source = (unsigned short *) src;
		unsigned short *dest = (unsigned short *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}
	else
	{
		const unsigned int *source = (unsigned int *) src;
		unsigned int *dest = (unsigned int *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}

}

////////////////////////////////////////////////////////////////////////
// MIRRORHORIZ - mirrors the image along horizontal axis
//
// Parameters:
//  dst          - destination array
//  src          - source array
//  widthbefore  - width of the source image in pixels
//  heightbefore - height of the source image in pixels
//
void COrientationHelper::MIRRORHORIZ( char *dst, char *src, int widthbefore, int heightbefore, int physbpp )
{
	if (physbpp <= 8)
	{
		unsigned char *source = (unsigned char *) src;
		unsigned char *dest = (unsigned char *) dst;

		for (int y=0;y<heightbefore;y++)
		{
			unsigned char *srcline = source + (heightbefore-y-1)*widthbefore;
			unsigned char *dstline = dest + (y*widthbefore);
			memman.memcpysw( (char *) dstline, (char *) srcline, widthbefore*sizeof(unsigned char));
		}
	}
	else if (physbpp <= 16)
	{
		unsigned short *source = (unsigned short *) src;
		unsigned short *dest = (unsigned short *) dst;

		for (int y=0;y<heightbefore;y++)
		{
			unsigned short *srcline = source + (heightbefore-y-1)*widthbefore;
			unsigned short *dstline = dest + (y*widthbefore);
			memman.memcpysw( (char *) dstline, (char *) srcline, widthbefore*sizeof(unsigned short));
		}
	}
	else
	{
		unsigned int *source = (unsigned int *) src;
		unsigned int *dest = (unsigned int *) dst;

		for (int y=0;y<heightbefore;y++)
		{
			unsigned int *srcline = source + (heightbefore-y-1)*widthbefore;
			unsigned int *dstline = dest + (y*widthbefore);
			memman.memcpysw( (char *) dstline, (char *) srcline, widthbefore*sizeof(unsigned int));
		}
	}

}

////////////////////////////////////////////////////////////////////////
// MIRRORVERT - mirrors the image along vertical axis
//
// Parameters:
//  dst          - destination array
//  src          - source array
//  widthbefore  - width of the source image in pixels
//  heightbefore - height of the source image in pixels
//
void COrientationHelper::MIRRORVERT( char *dst, char *src, int widthbefore, int heightbefore, int physbpp )
{
	if (physbpp <= 8)
	{
		unsigned char *source = (unsigned char *) src;
		unsigned char *dest = (unsigned char *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}
	else if (physbpp <= 16)
	{
		unsigned short *source = (unsigned short *) src;
		unsigned short *dest = (unsigned short *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}
	else
	{
		unsigned int *source = (unsigned int *) src;
		unsigned int *dest = (unsigned int *) dst;

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}

}

void COrientationHelper::ROT90ANDMIRRHORIZ(char *dst, char *src, int widthbefore, int heightbefore, int physbpp)
{
	if (physbpp <= 8)
	{
		unsigned char *source = (unsigned char *) src;
		unsigned char *dest = (unsigned char *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}
	else if (physbpp <= 16)
	{
		unsigned short *source = (unsigned short *) src;
		unsigned short *dest = (unsigned short *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}
	else
	{
		unsigned int *source = (unsigned int *) src;
		unsigned int *dest = (unsigned int *) dst;

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}

}
