
#include <winsock2.h>
#ifdef _WINDOWS_
#undef _WINDOWS_
#endif
#include "Stdafx.h"
#include "headers.h"
#include "SSyncSocket.h"

#undef __SW_DEBUG__
//#define __SW_DEBUG__

/////////////////////////////////////////////////////////////////////////////
// SSyncSocket - synchronous socket class
//
// This class is required for network connection support. It is an encapsulation 
// of the windows socket API. The whole class is synchronous (blocking). Each 
// function blocks the execution flow until the specific call can be completed.
//
// It is a good advise to use threads in order not to block the main application's
// execution flow. For example there can be listen, send and receive threads.
//
// Each function (except Send/Receive and the error handling functions) returns a 
// BOOLean error code. TRUE means that the function call completed successfully, while
// FALSE means that there was an error. The specific error code and the error string
// can be obtained by calling GetLastError() and GetLastErrorString().
//

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// construction: just invalidate the member socket handle
//
SSyncSocket::SSyncSocket():
	the_socket(INVALID_SOCKET)
{
}

//////////////////////////////////////////////////////////////////////////////
//
// destruction: if there is a working socket, shut down and close it
//
SSyncSocket::~SSyncSocket()
{
	if(the_socket!=INVALID_SOCKET){
		ShutDown();
		::closesocket(the_socket);
		the_socket=INVALID_SOCKET;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// This function binds this socket to a specific address of type SOCKADDR
//
BOOL SSyncSocket::Bind( const SOCKADDR * lpSockAddr, int nSockAddrLen )
{
    int ret;

	ret = ::bind(the_socket, lpSockAddr, nSockAddrLen);
	if (ret == 0) return( TRUE ); else { m_iLastError = 10; return( FALSE ); }
}

//////////////////////////////////////////////////////////////////////////////
//
// This function binds the socket to a specific port, and to a specific address.
// The second parameter is optional, that means you do not have to specify a socket 
// address. Just pass NULL and the socket will be bound to INADDR_ANY (any internet 
// address).
//
// The nSocketPort parameter is an integer in the range of 0..65535 (port range)
// The lpszSocketAddress parameter is a string which contains the IP address (e.g.
// 192.168.0.1)
//
// The function returns true if the socket can be bound. Otherwise it returns an 
// error. You can check GetLastError/+String to find out what went wrong.
//
BOOL SSyncSocket::Bind( UINT nSocketPort, LPCTSTR lpszSocketAddress )
{
    struct sockaddr_in A;
	int ret;

	if (lpszSocketAddress == NULL)
	{
       A.sin_family=AF_INET;
       A.sin_port=htons(nSocketPort);
       A.sin_addr.s_addr=INADDR_ANY;

       ret = bind( the_socket, (const struct sockaddr *) &A, sizeof(A)); 
	}
	 else
	{
    	A.sin_family=AF_INET;
		A.sin_port=htons(nSocketPort);
		A.sin_addr.s_addr= inet_addr( lpszSocketAddress );
		 
		ret = bind(the_socket, (const struct sockaddr *) &A, sizeof(A));
	}

	if (ret == 0) return( TRUE ); else { m_iLastError = 11; return( FALSE ); }
}

//////////////////////////////////////////////////////////////////////////////
//
// This function returns the IP adress and the port of the peer computer the 
// socket is connected to. The port is returned via a reference and the constructed 
// IP address is copied to the memory location rPeerAddress points to. You have to
// make sure that there are at least 16 bytes of storage space.
//
BOOL SSyncSocket::GetPeerName( LPSTR rPeerAddress, UINT & rPeerPort)
{
 struct sockaddr_in C;
 int length;
 char ipaddr[16]="\0";
 char temp[5];

 length = sizeof(C);
	
 int ret = getpeername( the_socket, (struct sockaddr *) &C, &length);

 if (ret==0)
 {
	 rPeerPort = ntohs(C.sin_port);

	 _itoa( (int) C.sin_addr.S_un.S_un_b.s_b1, temp, 10);
     strcat(ipaddr, temp);
	 strcat(ipaddr, ".");

	 _itoa( (int) C.sin_addr.S_un.S_un_b.s_b2, temp, 10);
     strcat(ipaddr, temp);
	 strcat(ipaddr, ".");

	 _itoa( (int) C.sin_addr.S_un.S_un_b.s_b3, temp, 10);
     strcat(ipaddr, temp);
	 strcat(ipaddr, ".");

	 _itoa( (int) C.sin_addr.S_un.S_un_b.s_b4, temp, 10);
     strcat(ipaddr, temp);
 
	 strcpy((char *) rPeerAddress, (const char *) &ipaddr);
     
     return( TRUE );
 } 
  else { m_iLastError = 12; return( FALSE ); }
}

//////////////////////////////////////////////////////////////////////////////
//
// This function returns the IP address and port of the peer computer the 
// socket is connected to. But instead of decoding the data (as the other
// GetPeerName function) this function just returns the SOCKADDR structure.
// 
// The programmer has to manually decode the information delivered in this 
// structure.
//
BOOL SSyncSocket::GetPeerName( SOCKADDR *lpSockAddr, int *lpSockAddrLen)
{
 int ret = getpeername( the_socket, lpSockAddr, lpSockAddrLen);

 if (ret==0) return( TRUE ); else { m_iLastError = 13; return( FALSE ); }
}

//////////////////////////////////////////////////////////////////////////////
//
// This function connects to the specific host/IP adress and port of the remote
// computer. For the lpszHostAddress parameter, you can choose whether you 
// enter the IP address or a host name. The nHostPort can be an integer from 
// 0..65535 .
//
BOOL SSyncSocket::Connect( LPCTSTR lpszHostAddress, UINT nHostPort )
{
 struct sockaddr_in C;

 // find out whether the lpszHostAddress is a dotted IP address
 C.sin_family = AF_INET;
 C.sin_port = htons(nHostPort);
 C.sin_addr.s_addr= inet_addr( lpszHostAddress );

 if (C.sin_addr.s_addr == INADDR_NONE) // aha, no dotted IP address, 
 {                                     // so look up hostname
  // pcHost isn't a dotted IP address, so send it to the resolver
  hostent* he = gethostbyname(lpszHostAddress);
  if (he == 0) {
		// No one knows who lpszHostAddress is so fail connect
    m_iLastError = 15;
	return FALSE;
  }

  // host found, convert the address into the appropriate structure
  C.sin_addr.s_addr = *((unsigned long*)(he->h_addr));
 }

 // issue the connect request
 int ret = connect(the_socket, (const struct sockaddr *) &C, sizeof(C));

 if (ret==0 ) return( TRUE ); else { m_iLastError = 14; return( FALSE ); }
}


//////////////////////////////////////////////////////////////////////////////
//
// This function connects to the specific SOCKADDRP adress and port of the remote
// computer. This function is different from the other connect function. Instead of 
// passing a logical value (like hostname and port), you have to pass a SOCKADDR 
// structure.
//
BOOL SSyncSocket::Connect( const SOCKADDR *lpSockAddr, int nSockAddrLen )
{
 int ret = connect(the_socket, lpSockAddr, nSockAddrLen); // connect to peer

 if (ret==0) return( TRUE ); else { m_iLastError = 16; return( FALSE ); }
}

//////////////////////////////////////////////////////////////////////////////
//
// This function closes a socket. It first shuts down sending on the socket, 
// then closes it. After closing the socket you have to call Create() again in 
// order to properly use it.
//
void SSyncSocket::Close()
{
    ShutDown();

	closesocket(the_socket);

	the_socket = INVALID_SOCKET;
}

//////////////////////////////////////////////////////////////////////////////
//
// This function creates a socket. All three parameters have default values 
// and thus, are optional. If you want to create a listen socket, you should
// at least use the nSocketPort parameter to bind it to your favourite port.
//
// Parameters:
//   nSocketPort       - port value (0..65535), default = 0
//   nSocketType       - SW_SOCK_STREAM : streaming oriented socket, default
//                       SOCK_DGRAM     : datagram socket
//   lpszSocketAddress - the address to bind the socket to, default = NULL
//
BOOL SSyncSocket::Create( UINT nSocketPort, int nSocketType, LPCTSTR lpszSocketAddress)
{
  the_socket = socket( AF_INET, nSocketType, 0 );

  if (the_socket != INVALID_SOCKET)
  {
		  if (Bind( nSocketPort, lpszSocketAddress) == TRUE)
		  {
				  return TRUE;
		  }
           else
          {
           m_iLastError = 4;
          }
  }
   else
  {
   m_iLastError = 2;
  }
  
  Close();

  return( FALSE );
}

////////////////////////////////////////////////////////////////////
// return the last error from underlying socket implementation
//
int SSyncSocket::GetLastError() 
{
	return WSAGetLastError();
}

////////////////////////////////////////////////////////////////////
// 
// This function shuts down the socket connection. You can specify the 
// how parameter. You can only shut down sending, receiving, or both 
// directions.
//
// Parameter:
//   how - SD_SEND     - shuts down sending on the socket
//       - SD_RECEIVE  - shuts down receiving on the socket
//		 - SD_BOTH     - shuts down sending and receiving on the socket
//
BOOL SSyncSocket::ShutDown(int how/* = SD_SEND*/)
{
	if(::shutdown(the_socket, how) == SOCKET_ERROR)
       {
           m_iLastError= 6;
		return FALSE;
       }
	else
		return TRUE;
}

////////////////////////////////////////////////////////////////////
// 
// This function puts the socket into listening mode. It listens for 
// incoming connections. This function blocks until there is a 
// connection attempt. You should use it in a separate thread in order
// not to block the main execution flow.
//
// Parameter:
//   backlog  - the number of simultaneous connection attempts the socket buffers
//              can not be greater than SOMAXCONN
//
BOOL SSyncSocket::Listen(int backlog/* = SOMAXCONN*/)
{
	if(::listen(the_socket, backlog) == SOCKET_ERROR)
       {
           m_iLastError = 7;
		return FALSE;
       }
	else
		return TRUE;
}

///////////////////////////////////////////////////////////////////
//
// This private function attaches a socket handle to the class. A socket
// which might be attached to the class before is closed. When the socket
// which is passed to the function is an invalid one, this function returns
// an error.
//
bool SSyncSocket::attach(SOCKET s)
{
	if(s!=INVALID_SOCKET)
    {
	    Close();
		the_socket=s;
 		return true;

	}

    m_iLastError = 8;  
	return false;
}

///////////////////////////////////////////////////////////////////
//
// This function has to be called after a the listen() function returns.
// The resulting socket of the connection attempt will be transferred into
// the rConnectedSocket class. This function blocks until there is a pending
// connection.
//
BOOL SSyncSocket::Accept( SSyncSocket &rConnectedSocket, SOCKADDR *lpSockAddr, int *lpSockAddrLen )
{
  SOCKET new_socket;
  struct sockaddr_in B;
  int length;

  length = sizeof(B);

  // accept the incoming connection attempt
  new_socket = ::WSAAccept(the_socket, (struct sockaddr *) &B, &length, NULL, 0 );
  if (new_socket == INVALID_SOCKET)
  {
   m_iLastError = 9;
   return FALSE;
  }

  // attach the new socket to the passed socket reference
  rConnectedSocket.attach(new_socket);

  // if the programmer has passed return variables, then fill them in
  if (lpSockAddrLen != NULL) *lpSockAddrLen = length;
  if( lpSockAddr != NULL) memcpy( (void*) lpSockAddr, (void *) &B, sizeof(B));

  return TRUE;
}
///////////////////////////////////////////////////////////////////
//
// This function receives nBufLen bytes from the socket. Unlike the 
// low-level recv function this one does not read up to nBufLen bytes,
// but blocks until exactly nBufLen bytes have read. This feature needs
// to be there for proper reading of some data structures out of 
// the socket.
//
// Parameters:
//  void *lpBuf - memory location where the read bytes are stored
//  int nBufLen - number of bytes to read
//  nFlags      - is directly passed to the recv-function, see
//                MSDN documentation - optional parameter
//
// Return value:
//   SOCKET_ERROR (-1) : there was an error while receiving from the socket
//   n ; n != -1       : number of bytes read from the socket
int SSyncSocket::Receive(void *lpBuf, int nBufLen, int nFlags)
{
    int total=0;
    int total_max=nBufLen;
    char *moveptr = (char *) lpBuf;

    while( total < total_max)
    {
     int numread = 4000;
     if (total_max-total < numread) numread = total_max-total;
   
     int len = ::recv(the_socket, moveptr, numread, nFlags);

     if (len != SOCKET_ERROR && len != 0)
     {
      total += len;
      moveptr += len;
     }
      else
     {
      int ret = ::WSAGetLastError();
      if (ret != WSAEINTR) 
	  {
		  return( SOCKET_ERROR );
	  }
     }
    }

	return total;
}

///////////////////////////////////////////////////////////////////
//
// This function sends nBufLen bytes from memory location lpBuf to 
// the peer computer. This function blocks until all data has been 
// sent.
//
// Parameters:
//  void *lpBuf - memory location where the to be written bytes are stored
//  int nBufLen - number of bytes to write
//  nFlags      - is directly passed to the send-function, see
//                MSDN documentation - optional parameter
//
// Return value:
//   SOCKET_ERROR (-1) : there was an error while sending data to the socket
//   n ; n != -1       : number of bytes written to the socket
//
int SSyncSocket::Send( const void *lpBuf, int nBufLen, int nFlags)
{
	int send = 0;
	int len;
	int lasterror = 0;
	while( send != 1)
	{
	 len = ::send(the_socket, (const char *) lpBuf, nBufLen, nFlags);
	 if (len == SOCKET_ERROR) lasterror = WSAGetLastError(); else lasterror = 0;
	 if (lasterror != WSAEINTR)
		 send = 1;
	}

	return( len );
}

///////////////////////////////////////////////////////////////////
//
// This function retrieves a human readable string of the last error.
// An even more specific error code can be obtained by GetLastError()
// which returns the last error code on the Windows socket 
// implementation. For further information on Windows sockets error 
// codes please consult MSDN Documentation.
//
// Parameter:
//  CString &msg - by reference parameter in which the error description 
//                 string is returned.
//
// If you compile this class with #define __SW_DEBUG__ symbol, this function
// also prints out a messagebox with the error description.
//  
void SSyncSocket::GetLastErrorString(CString &msg)
{
 switch( m_iLastError )
 {
  case  0: break;

  case  2: msg.Format("Invalid socket handle. (CRE/%d)", m_iLastError);
           break;
  case  4: msg.Format("Could not bind socket (CRE/%d)", m_iLastError);
           break;
  case  6: msg.Format("Could not shut down socket. (SHD/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case  7: msg.Format("Could not initialize listensocket. (LIS/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case  8: msg.Format("Could not attach socket. (ATT/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case  9: msg.Format("WSAAccept returned invalid socket. (ACC/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 10: msg.Format("Error BIND socket. (BND/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 11: msg.Format("Error BIND socket. (BN1/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 12: msg.Format("Error in GetPeerName. (GPN/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 13: msg.Format("Error in GetPeerName. (GP1/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 14: msg.Format("Failed connect. (CON/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 15: msg.Format("Failed gethostbyname. (CON/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;
  case 16: msg.Format("Failed connect. (CO1/%d) (WSA=%xh)", m_iLastError, GetLastError());
           break;

  default: msg.Format("Unknown error: %d (maybe WSAERR=%xh)", m_iLastError, GetLastError() );
           break;

 }

#ifdef __SW_DEBUG__

 if (m_iLastError != 0)
 {
  AfxMessageBox((LPCTSTR) msg, MB_OK | MB_ICONINFORMATION );
 }

#endif
}

