// Stat.cpp: implementation of the CStat class.
//
// Initially written by Velizar Miltchev (DESY Zeuthen) miltchev@ifh.de
// some additional code and modifications for Grabber server 
// by Stefan Weisse sweisse@ifh.de
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "headers.h"
#include "Stat.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
// Stat class from Videoclnt without background subtraction, normalizing etc.
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// 
// The constructor of CStat does some basic initialization. Please note that in this version on
// the grabber server, no background image is subtracted, no area of interest is normalized and 
// no image is (maybe) x-ray filtered. Only the projections are calculated. 
//
// Parameters:
//  data     - pointer to the to be analysed image (must contain at least (ImWi*ImHi bytes)
//  ImWi     - width of the image
//  ImHi     - height of the image
//
CStat::CStat(unsigned char* data,unsigned int ImWi,unsigned int ImHi)
{
    // initialize member variables through parameters
	array=data;
	BackGr = NULL;
	ImageWidth=ImWi;
	ImageHeight=ImHi;

    // initialize mean and rms values to default value
	XMEAN=0;
	YMEAN=0;
	XRMS=0;
	YRMS=0;

    // allocate needed distribution arrays 
	YDIST= (double *) memman.sw_malloc( sizeof(double)*ImageHeight );
	XDIST= (double *) memman.sw_malloc( sizeof(double)*ImageWidth );
    YDISTi = (unsigned int *) memman.sw_malloc( sizeof(unsigned int)*ImageHeight );
    XDISTi = (unsigned int *) memman.sw_malloc( sizeof(unsigned int)*ImageWidth );

	if (XDIST != NULL && YDIST != NULL && YDISTi != NULL && XDISTi != NULL)
	{
      // create clip values
      up =0;
	  down = ImHi - 1;
	  left = 0;
	  right = ImWi - 1;


      // calculate distribution arrays
      CalcXYDistSW(array);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// In the destructor just the dynamically allocated variables are freed.
//
CStat::~CStat()
{
    // free floating point distribution arrays	
	if (XDIST != NULL) memman.sw_free( (void *) XDIST );
	if (YDIST != NULL) memman.sw_free( (void *) YDIST );

    // free integer temporary distribution arrays
	if (XDISTi != NULL) memman.sw_free( (void *) XDISTi );
	if (YDISTi != NULL) memman.sw_free( (void *) YDISTi );
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// In this function, the spot center coordinates and size of the spot is calculated. After 
// calling this function, one can get the calculated value by calling functions like X_RMS(),
// Y_RMS, XM() and YM().
//
void CStat::START()
{
	if (XDIST != NULL && YDIST != NULL)
	{
		XMEAN=CalcXMEAN(XDIST);
		YMEAN=CalcYMEAN(YDIST);
		XRMS=CalcXRMS(XDIST,XMEAN);
		YRMS=CalcYRMS(YDIST,YMEAN);
	}
}

/* does not normalize the image, just calculate
void CStat::NormalizeImage(unsigned char *datablock, unsigned char *outdata)
{
 int xlen = right - left;
 int ylen = down - up;

 if (xlen < 1 || ylen < 1) return;

 for (int y=up;y<down;y++)
 {
  sum = 0;
  for (int x=left;x<right;x++)
  {
   sum += (unsigned int) data[y*ImageWidth+x];
   XDISTi[x] += (unsigned int) data[y*ImageWidth+x];
  }
  YDISTi[y] = sum;
 }
}
*/

//////////////////////////////////////////////////////////////////////////////////////////
//
// This function calculates the XDIST and YDIST distribution arrays. These arrays are 
// needed for proper calculation of the mean and rms values (spot center coordinates and 
// spot size).
//
void CStat::CalcXYDistSW(unsigned char *data)
{
 ZeroMemory(YDIST, ImageHeight*sizeof(double));
 ZeroMemory(XDIST, ImageWidth*sizeof(double));
 ZeroMemory(XDISTi, ImageWidth*sizeof(unsigned int));
 ZeroMemory(YDISTi, ImageHeight*sizeof(unsigned int));
 
 int xlen = right - left + 1;
 int ylen = down - up + 1;

 register unsigned int sum = 0;
 register unsigned int xstart = up*ImageWidth+left;
 register unsigned int xadd = (ImageWidth-1) - (right - left);
 
 for (register unsigned int y=up;y<=down;y++)
 {
  sum = 0;
  for (register unsigned int x=left;x<=right;x++)
  {
   register unsigned int p;
   p = (unsigned int) data[xstart];
   // calculation of nDISTi
   sum += (unsigned int) p;
   XDISTi[x] += (unsigned int) p;
 
   xstart++;
  }
  xstart += xadd;
  YDISTi[y] = sum;
 }

 sum = 1;
 for (register int x=(int)left;x<=(int)right;x++)
 {
  sum += XDISTi[x];
 }

 if (sum != 1) sum--;

 double ZEROh = 1.0 * (double)(sum==1)/ ((double) (right-left+1));
 double ZEROv = 1.0 * (double)(sum==1)/ ((double) (down-up+1));

 // scale xdist accordingly
 for (x=(int)left;x<=(int)right;x++)
 {
  XDIST[x] = (ZEROh+(double)XDISTi[x])/((double) sum);
 }

 // scale ydist accordingly
 for (y=up;y<=down;y++)
 {
  YDIST[y] = (ZEROv+(double)YDISTi[y])/((double) sum);
 }

}

////////////////////////////////////////////////////////////////////////////
//
// This function is just left into for backward compatibility. It is not used
// any more. It was replaced by CalcXYDistSW().
//
void CStat::CalcXDIST(unsigned char* data)
{
	MaxVal=1;
	ZeroMemory(XDIST,ImageWidth*sizeof(double));
	double sumy,sum=1;
	unsigned char local,d,b;
	for(unsigned int i=left;i<=right;i++)
	{
		sumy=0;
		for(unsigned int j=up;j<=down;j++)
		{
			d=data[i+j*ImageWidth],b=BackGr[i+j*ImageWidth];
			sumy+=(local=(data[i+j*ImageWidth]=(d>b)?d-b:0));
			MaxVal=(local>MaxVal)?local:MaxVal;
		}
		XDIST[i]=sumy;
		sum+=sumy;
	}
	for(i=left;i<=right;i++)XDIST[i]/=sum;
}

////////////////////////////////////////////////////////////////////////////
//
// This function is just left into for backward compatibility. It is not used
// any more. It was replaced by CalcXYDistSW().
//
void CStat::CalcYDIST(unsigned char* data)
{
	ZeroMemory(YDIST,ImageHeight*sizeof(double));
	double sumx,sum=1,coef=255/MaxVal;
	for(unsigned int j=up;j<=down;j++)
	{
		sumx=0;
		for(unsigned int i=left;i<=right;i++)
		{
            double dataval = data[i+j*ImageWidth];
			dataval *= coef;
			data[i+j*ImageWidth] = (unsigned char) dataval;
			sumx += dataval;
		}
		YDIST[j]=sumx;
		sum+=sumx;
	}
	for(j=up;j<=down;j++)YDIST[j]/=sum;
}


////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot center on the x axis out of the 
// distribution array that is passed as 'dist'.
//
double  CStat::CalcXMEAN(double* dist)
{
	double RV=0;
	for(unsigned int i=left;i<=right;i++)
		RV+=(double)i*dist[i];
	return RV;
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot center on the Y axis out of the 
// distribution array that is passed as 'dist'.
//
double  CStat::CalcYMEAN(double* dist)
{
	double RV=0;
	for(unsigned int i=up;i<=down;i++)
		RV+=(double)i*dist[i];
	return RV;
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot size on the X axis out of the 
// distribution array that is passed as 'dist' and the x mean value passed
// as 'xm'.
//
double  CStat::CalcXRMS(double* dist,double xm)
{
	double RV=0;
	int l=right-left;
	for(unsigned int i=left;i<=right;i++)RV+=(i-xm)*(i-xm)*dist[i];
	//FitAndSpline(&XDIST[left],&XDIST[left],l,1+9*(l>10));
	return sqrt(RV);
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot size on the Y axis out of the 
// distribution array that is passed as 'dist' and the y mean value passed
// as 'ym'.
//
double  CStat::CalcYRMS(double* dist,double ym)
{
	double RV=0;
	int l=down-up;
	for(unsigned int i=up;i<=down;i++)RV+=(i-ym)*(i-ym)*dist[i];
	//FitAndSpline(YDIST,YDIST,l,1+9*(l>10));
	return sqrt(RV);
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the XDIST distribution array. Plase note that you 
// obtain a pointer to an internal array. So after destruction of this class the 
// array is no longer valid !
//
double*  CStat::X_DIST()
{
	return XDIST;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the YDIST distribution array. Plase note that you 
// obtain a pointer to an internal array. So after destruction of this class the 
// array is no longer valid !
//
double*  CStat::Y_DIST()
{
	return YDIST;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot size in x dimension. You should only call
// this function after calling START().
//
double  CStat::X_RMS()
{
	return XRMS;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot size in y dimension. You should only call
// this function after calling START().
//
double  CStat::Y_RMS()
{
	return YRMS;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot center in x dimension. You should only call
// this function after calling START().
//
double  CStat::XM()
{
	return XMEAN;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot center in y dimension. You should only call
// this function after calling START().
//
double  CStat::YM()
{
	return YMEAN;
}


