// CFourierStat.cpp: implementation of the CCFourierStat class.
//
// Initially written by Velizar Miltchev (DESY Zeuthen) miltchev@ifh.de
// some additional code and modifications for Grabber server 
// by Stefan Weisse sweisse@ifh.de
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "headers.h"
#include "FourierStat.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// 
// The constructor of CFourierStat does some basic initialization. In this special version for 
// the grabber server: the background image NOT is subtracted, NO area of interest is normalized and
// the image is NOT x-ray filtered. Only the projections are calculated.
//
// Parameters:
//  data     - pointer to the to be analysed images (must contain at least (ImWi*ImHi bytes)
//  deg      - degree
//  recdep   - recursion depth of fourier analysis
//  ImWi     - width of the image
//  ImHi     - height of the image
//
CFourierStat::CFourierStat(unsigned char* data, unsigned short deg,unsigned short recdep,unsigned int ImWi,unsigned int ImHi)
		:CStat(data,ImWi,ImHi)
{
	// initialize fourier parameters
	degree=deg;
	RecursDepth=recdep;
}


////////////////////////////////////////////////////////////////////////////////////////////
//
// Destruction is empty
//
//
CFourierStat::~CFourierStat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the A coefficent needed by fourier analysis. The function does
// not use the member variables of the class, but the parameters.
//
// Parameters:
//  DistArray - distribution array to analyse
//  len       - length of the distribution array
//  deg       - number of degrees
//
// Return value:
//  Coefficent array that was dynamically allocated
//
double* CFourierStat::CalcAcoef(double* DistArray,int len, unsigned short deg)
{
	double locl=len;
	double wavenum=2*pi/(locl+1e-10);
	double* coefarr= (double *) memman.sw_malloc( sizeof(double)*degree );
	if (coefarr == NULL) return NULL;
	ZeroMemory(coefarr,degree*sizeof(double));
	for( short i=0;i<degree;i++)
		for(int j=0;j<=len;j++)
			coefarr[i]+=2*DistArray[j]*cos(((double)(i*j))*wavenum)/(1e-10+(double)len);
	return coefarr;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the B coefficent needed by fourier analysis. The function does
// not use the member variables of the class, but the parameters.
//
// Parameters:
//  DistArray - distribution array to analyse
//  len       - length of the distribution array
//  deg       - number of degrees
//
// Return value:
//  Coefficent array that was dynamically allocated. Make sure to deallocate it.
//
double* CFourierStat::CalcBcoef(double* DistArray,int len, unsigned short deg)
{	

	double locl=len;
	double wavenum=2*pi/(locl+1e-10);
	double* coefarr= (double *) memman.sw_malloc( sizeof(double)*degree );
	if (coefarr == NULL) return NULL;
	ZeroMemory(coefarr,degree*sizeof(double));
	for( short i=0;i<degree;i++)
		for(int j=0;j<=len;j++)
			coefarr[i]+=2*DistArray[j]*sin(((double)(i*j))*wavenum)/(1e-10+(double)len);
	return coefarr;
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function recalculates the XDIST distribution array out of the base data found in XDIST
// distribution array. The recalculation is needed because the fourier analysis needs a 
// different distribution array than calculations done in CStat.
//
// Parameters:
//  (XDIST)  - no real parameter, but member variable which needs to be properly initialized
//  locleft  - starting index inside XDIST array 
//  locright - ending index inside XDIST array 
//
// Return value:
//  ReXDist  - calculated Fourier distribution array. This array is dynamically allocated so 
//             make sure that it gets properly deallocated.
//
double* CFourierStat::ReXDist(int locleft, int locright)
{
	double len=((double)(locright-locleft));len+=1e-10;
	double* ReDist= (double *) memman.sw_malloc( sizeof(double)*ImageWidth );
	double* coefA=CalcAcoef(&XDIST[locleft],(int)len,degree);
	double* coefB=CalcBcoef(&XDIST[locleft],(int)len,degree);
	if (coefA == NULL || coefB == NULL || ReDist == NULL) 
	{
     	if (coefA != NULL) memman.sw_free( (void *) coefA );
  	    if (coefB != NULL) memman.sw_free( (void *) coefB );
		if (ReDist != NULL) memman.sw_free( (void *) ReDist );
		return NULL;
	}
	double wavenum=2*pi/len;
	double sum=1e-10;
	for(int i=locleft;i<=locright;i++)
	{
		ReDist[i]=coefA[0]/2;
		for(int j=1;j<degree;j++)ReDist[i]+=coefA[j]*cos(((double)(j*(i-locleft)))*wavenum)+
			                                coefB[j]*sin(((double)(j*(i-locleft)))*wavenum);
		ReDist[i]*=(ReDist[i]>0);
		sum+=(ReDist[i]+=1e-10);
	}
	for(i=locleft;i<=locright;i++)ReDist[i]/=sum;
	if (coefA != NULL) memman.sw_free( (void *) coefA );
	if (coefB != NULL) memman.sw_free( (void *) coefB );
	return ReDist;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function recalculates the YDIST distribution array out of the base data found in YDIST
// distribution array. The recalculation is needed because the fourier analysis needs a 
// different distribution array than calculations done in CStat.
//
// Parameters:
//  (YDIST)  - no real parameter, but member variable which needs to be properly initialized
//  locup    - starting index inside YDIST array 
//  locdown  - ending index inside YDIST array 
//
// Return value:
//  ReYDist  - calculated Fourier distribution array. This array is dynamically allocated so 
//             make sure that it gets properly deallocated.
//
double* CFourierStat::ReYDist(int locup, int locdown)
{
	double len=((double)(locdown-locup));len+=+1e-10;
	double* ReDist=(double *) memman.sw_malloc( sizeof(double)*ImageHeight );
	double* coefA=CalcAcoef(&YDIST[locup],(int)len,degree);
	double* coefB=CalcBcoef(&YDIST[locup],(int)len,degree);
	if (coefA == NULL || coefB == NULL || ReDist == NULL) 
	{
     	if (coefA != NULL) memman.sw_free( (void *) coefA );
  	    if (coefB != NULL) memman.sw_free( (void *) coefB );
		if (ReDist != NULL) memman.sw_free( (void *) ReDist );
		return NULL;
	}
	double wavenum=2*pi/len;
	double sum=1e-10;
	for(int i=locup;i<=locdown;i++)
	{
		ReDist[i]=coefA[0]/2;
		for(int j=1;j<degree;j++)ReDist[i]+=coefA[j]*cos(((double)(j*(i-locup)))*wavenum)+
			                                coefB[j]*sin(((double)(j*(i-locup)))*wavenum);
		ReDist[i]*=(ReDist[i]>0);
		sum+=(ReDist[i]+=1e-10);
	}
	for(i=locup;i<=locdown;i++)ReDist[i]/=sum;
	if (coefA != NULL) memman.sw_free( (void *) coefA );
	if (coefB != NULL) memman.sw_free( (void *) coefB );
	return ReDist;
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the XMEAN (spot center in x dimension) variable out of the passed
// parameters. The function calls itself recursively up to a depth of 'recnum'.
//
// Parameters:
//  dist     - distribution array to calculate x mean value
//  recnum   - recursion depth
//  locleft  - starting index inside dist array 
//  locright - ending index inside dist array 
//
// Return value:
//  xmean    - calculated x mean value (spot center position in x dimension)
//
double CFourierStat::CalcXMEAN( double* dist,unsigned short recnum, int locleft, int locright)
{
 double localxm=locleft,localxs=0;
 for(int i=locleft;i<=locright;i++)localxm=dist[i]>dist[(int)localxm]?i:localxm;
 if(recnum==1)return localxm;
 double sum=1e-10;
 for(i=locleft;i<=locright;i++)sum+=dist[i];
 for(i=locleft;i<=locright;i++)localxs+=(i-localxm)*(i-localxm)*dist[i]/sum;
 localxs=sqrt(localxs);
 int newleft=(int) (localxm-localxs>locleft?localxm-localxs:locleft);
 int newright=(int) (localxm+localxs<locright?localxm+localxs:locright);
 if(newright-newleft<2){newleft=locleft;newright=locright;}
 return CalcXMEAN(dist,recnum-1,newleft,newright);
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the YMEAN (spot center in y dimension) variable out of the passed
// parameters. The function calls itself recursively up to a depth of 'recnum'.
//
// Parameters:
//  dist     - distribution array to calculate y mean value
//  recnum   - recursion depth
//  locleft  - starting index inside dist array 
//  locright - ending index inside dist array 
//
// Return value:
//  ymean    - calculated y mean value (spot center position in y dimension)
//
double CFourierStat::CalcYMEAN( double* dist,unsigned short recnum, int locleft, int locright)
{
 double localxm=locleft,localxs=0;
 for(int i=locleft;i<=locright;i++)localxm=dist[i]>dist[(int)localxm]?i:localxm;
 if(recnum==1)return localxm;
 double sum=1e-10;
 for(i=locleft;i<=locright;i++)sum+=dist[i];
 for( i=locleft;i<=locright;i++)localxs+=(i-localxm)*(i-localxm)*dist[i]/sum;
 localxs=sqrt(localxs);
 int newleft=(int)(localxm-localxs>locleft?localxm-localxs:locleft);
 int newright=(int)(localxm+localxs<locright?localxm+localxs:locright);
 if(newright-newleft<2){newleft=locleft;newright=locright;}
 return CalcYMEAN(dist,recnum-1,newleft,newright);
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function does all calculations. It should be called right after the construction of
// the class and before calling any of the public member functions.
//
void CFourierStat::START()
{
    // calculate fourier distribution arrays
	FTXdist=ReXDist(left,right);
	FTYdist=ReYDist(up,down);
	if (FTXdist == NULL || FTYdist == NULL)
	{
    	if (FTXdist != NULL) memman.sw_free( (void *) FTXdist );
	    if (FTYdist != NULL) memman.sw_free( (void *) FTYdist );
		return;
	}

	// calc mean values (spot center coordinates)
	XMEAN=CalcXMEAN(FTXdist,RecursDepth,left,right);
	YMEAN=CalcYMEAN(FTYdist,RecursDepth,up,down); 

	// calculates size values (spot size in x and y dimension)
	XRMS=CalcXRMS(FTXdist,XMEAN);
	YRMS=CalcYRMS(FTYdist,YMEAN);

	// free up memory allocated while calculating fourier distribution arrays
	if (FTXdist != NULL) memman.sw_free( (void *) FTXdist );
	if (FTYdist != NULL) memman.sw_free( (void *) FTYdist );
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the spot size in x dimension. It does not use any member variables
// but the parameters instead.
//
// Parameters:
//  dist  - distribution array on which calculation is based
//  xm    - previously calculated x mean value
//
// Return value:
//  x rms - spot size in x coordinates
//
double CFourierStat::CalcXRMS(double *dist, double xm)
{
	double retv=0;
	double sum=1e-10;
	int lxr= (int) (2.78*CStat::CalcXRMS(XDIST,xm));
	int ll=(int)(xm-lxr>left?xm-lxr:left);
	int rl=(int)(xm+lxr<right?xm+lxr:right);
	left=(unsigned int)xm;right=(unsigned int)xm;
	for(int i=(int)xm;i>=ll;i--)left=(int)(dist[i]<dist[left]?i:left);
	for(i=(int)xm;i<=rl;i++)right=dist[i]<dist[right]?i:right;
	for(i=left;i<=(int)right;i++)sum+=XDIST[i];
	for(i=left;i<=(int)right;i++)retv+=(i-xm)*(i-xm)*XDIST[i]/sum;
	return sqrt(retv);
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the spot size in y dimension. It does not use any member variables
// but the parameters instead.
//
// Parameters:
//  dist  - distribution array on which calculation is based
//  ym    - previously calculated y mean value
//
// Return value:
//  y rms - spot size in y coordinates
//
double CFourierStat::CalcYRMS(double *dist, double ym)
{
	double retv=0;
	double sum=1e-10;
	int lxr=(int)(2.78*CStat::CalcYRMS(YDIST,ym));
	int ll=(int)(ym-lxr>up?ym-lxr:up);
	int rl=(int)(ym+lxr<down?ym+lxr:down);
	up=(unsigned int)ym;down=(unsigned int)ym;
	for(int i=(int)ym;i>=ll;i--)up=dist[i]<dist[up]?i:up;
	for(i=(int)ym;i<=rl;i++)down=dist[i]<dist[down]?i:down;
	for(i=up;i<=(int)down;i++)sum+=YDIST[i];
	for(i=up;i<=(int)down;i++)retv+=(i-ym)*(i-ym)*YDIST[i]/sum;
	return sqrt(retv);
}
