#include "platform.h"
#include "videofileload.h"
#include "getframes.h"

// external reference to decompression.c
extern void decompress_huffyuv( VOIDP compressed, VOIDP uncompressed, UINT32 size_uncompressed);

// in order to be compatible with lots of platforms and compiler settings (alignment), the 
// structure VIDHDR will be read in as an 88 byte bitstream and later converted to the proper
// structure, taking into account how the GrabServer2 put the structure together - there is no
// direct copy or assignment !
//
// easy reasonability checking will be done (kennung and bpp)
// -1		error in translation (bad stream input)
//  0		conversion succeeded
INT32 translate_charstream_into_vidhdr( VIDHDR *vidH, const UCHAR *stream)
{
	vidH->kennung = *((UINT32 *) stream);
	stream+=4;
	vidH->compression = *((UINT32 *) stream);
	stream+=4;
	vidH->len_total = *((UINT32 *) stream);
	stream+=4;
	vidH->len_data = *((UINT32 *) stream);
	stream+=4;
	vidH->len_fmthdr = *((UINT32 *) stream);
	stream+=4;
	vidH->frameno = *((UINT32 *) stream);
	stream+=4;
	vidH->timestamp.time = *((UINT32 *) stream);
	stream+=4;
	vidH->timestamp.millitm = *((UINT16 *) stream);
	stream+=2;
	vidH->timestamp.timezone = *((INT16 *) stream);
	stream+=2;
	vidH->timestamp.dstflag = *((INT16 *) stream);
	stream+=2;
	vidH->timestamp.reserved[0] = *((CHAR *) stream);
	stream+=1;
	vidH->timestamp.reserved[1] = *((CHAR *) stream);
	stream+=1;
	vidH->cameraport = *((UINT32 *) stream);
	stream+=4;
	vidH->widthframe = *((UINT32 *) stream);
	stream+=4;
	vidH->heightframe = *((UINT32 *) stream);
	stream+=4;
	vidH->compressed = *((UINT32 *) stream);
	stream+=4;
	memcpy(&vidH->reserved, stream, 12);
	stream+=12;
		
	vidH->scale = *((DOUBLE *) stream);
	stream+=8;
	vidH->bpp = *((UINT32 *) stream);
	stream+=4;
	vidH->bpp_effective = *((UINT32 *) stream);
	stream+=4;

	memcpy(&vidH->reserved1, stream, 8);
	
	//stream+=12;
	
	if (vidH->kennung != 0x11223344) return( -1 );
	if ((vidH->bpp != 8) && (vidH->bpp != 16) && (vidH->bpp != 32)) return( -1 );
		
	return 0;
}


// This is the inner-loop socket function that reads the raw data from the socket. It is written in such a 
// way that lower-level socket functions inside platform.c / platform.h are used in order to be a certain
// kind of platform independent.
//
//	Parameters:
//		_socket				handle for the socket
//		data				buffer to put the received data to
//		length				maximum length to read (in bytes, must not exceed total length of buffer)
//		length_returned		if not NULL, the total amount of bytes read will be put in here (in success case)
//		msec_timeout		timeout on the receive operation in milliseconds waiting for data to arrive
//
//	Return value:
//		1					success
//		2					timeout
//		any_other_val		low level socket error
//
INT32 receive_socket( socktype _socket, UCHAR *data, UINT32 length, UINT32 *length_returned, UINT32 msec_timeout )
{
	UINT32 total = 0;
	UINT32 total_max = length;
	UCHAR *moveptr = data;
	UINT32  errwait = 0;
	INT32  ecode;
	UINT32 msec10to = msec_timeout / 10;
	INT32  lret;

	while( total < total_max )
	{
		UINT32 numread = 4000;
		
		if ( (total_max-total) < numread ) numread = (total_max-total);
		
		lret = receive_socket_function(_socket, moveptr, numread);

		if (lret == 0 || lret == SOCKET_ERROR)
		{
			ecode = getlasterror();
			
			if (ecode == EAGAIN)
			{
				errwait++;
				
				// wait up to two seconds
				if (errwait == msec10to) return( 2 );// timeout
				
				sleep_milliseconds(10);
			}
			else
			{ 
				return( lret ); 
			}
		}
		else
		{
			moveptr += lret;
			total += lret;
		}
	}
	
	if (length_returned != NULL) *length_returned = total;
	
	return( 1 );
	// success
}

// macro that will be used to keep functions clean and tidy when
// reacting on error conditions
#define ERR(a)  \
		 { errcode = a; \
		goto err;}  \


// This function is used to downscale a vinf header containing multiple images to 
// a vinf header containing a combined, single image. The single image is created 
// out of the multiple images by applying a envelope function.
// 
// This function is helpful e.g. when converting BKC to BKG file.
//
//	Parameters:
//		[in/out]	vinf		VIDINFHDR containing multiple images [in]
//								VIDINFHDR containing the downscaled single image result [out]
//
//	Return values:
//		 0		done, finished successfully
//		-5		bpp not supported
//		-6		memory allocation error
//
//	Note: In any error case the vinf is not touched and the old data remains inside 
//		  (without corruption) !
//
INT32 downscale_envelope_function( struct VIDINFHDR *vinf )
{
	UINT32 *buffer32 = (UINT32 *) malloc( vinf->width*vinf->height*sizeof(UINT32) );
	INT32 num;
	UINT32 x;
	INT32 errcode=0;
	DOUBLE *local_framescales=NULL;
	VOIDP *local_framebits=NULL;
		
	if (buffer32 == NULL) ERR(-6);

	for (x=0;x<(vinf->width*vinf->height);x++)
	{
		buffer32[x] = 0;
	}
			
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->bpp <= 8)
		{
			UCHAR *imgbuf = (UCHAR *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = (UINT32) *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else if (vinf->bpp <= 16)
		{
			UINT16 *imgbuf = (UINT16 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = (UINT32) *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else if (vinf->bpp <= 32)
		{
			UINT32 *imgbuf = (UINT32 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else ERR( -5 );
	}
	
	// drop back buffer, adjust VIDINFHDR
	
	local_framescales = (DOUBLE *) malloc( sizeof(DOUBLE)*1 );
	if (local_framescales == NULL) ERR(-6);
		
	local_framebits = (VOIDP *) malloc( sizeof(VOIDP *)*1 );
	if (local_framebits == NULL) ERR(-6);
	
	local_framebits[0] = (VOIDP) malloc( (vinf->width*vinf->height*vinf->bpp)/8 );
	if (local_framebits[0] == NULL) ERR(-6);


	// free old memory buffers	
	
	if (vinf->framescales != NULL) free( vinf->framescales );
	
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->framebits[num] != NULL) free( (VOIDP) vinf->framebits[num] );
	}
	if (vinf->framebits != NULL) free( (VOIDP) vinf->framebits );
	
	// adjjust vinf header
	vinf->numframes = 1;	
	vinf->framescales = local_framescales;
	vinf->framebits = local_framebits;
	vinf->framescales[0] = vinf->scale;
	
	// update data	
		
	if (vinf->bpp <= 8)
	{
		UCHAR *imgbuf = (UCHAR *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UCHAR) buffer32[x];
		}
	}
	else if (vinf->bpp <= 16)
	{
		UINT16 *imgbuf = (UINT16 *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT16) buffer32[x];
		}
	}
	else if (vinf->bpp <= 32)
	{
		UINT32 *imgbuf = (UINT32 *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = buffer32[x];				
		}
	}
	
	
err:
	if (errcode != 0)
	{
  		if (local_framescales != NULL) free( (VOIDP) local_framescales );
  		if (local_framebits != NULL)
  		{
  			if (local_framebits[0] != NULL) free( (VOIDP) local_framebits[0] );
  			free( (VOIDP) local_framebits );
  		}
	}
	
	free( (VOIDP) buffer32 );

	return errcode;
}

// This function is used to downscale a vinf header containing multiple images to 
// a vinf header containing a combined, single image. The single image is created 
// out of the multiple images by applying an 'averaging' function.
// 
// This function is helpful e.g. when converting BKC to BKG file.
//
//	Parameters:
//		[in/out]	vinf		VIDINFHDR containing multiple images [in]
//								VIDINFHDR containing the downscaled single image result [out]
//
//	Return values:
//		 0		done, finished successfully
//		-5		bpp not supported
//		-6		memory allocation error
//
//	Note: In any error case the vinf is not touched and the old data remains inside 
//		  (without corruption) !
//
INT32 downscale_average_function( struct VIDINFHDR *vinf )
{
	DOUBLE *bufferD = (DOUBLE *) malloc( vinf->width*vinf->height*sizeof(DOUBLE) );
	INT32 num;
	UINT32 x;
	INT32 errcode=0;
	DOUBLE *local_framescales=NULL;
	VOIDP *local_framebits=NULL;
	DOUBLE old_num_frames = (DOUBLE) vinf->numframes;
	
	if (bufferD == NULL) ERR(-6);

	for (x=0;x<(vinf->width*vinf->height);x++)
	{
		bufferD[x] = 0.0;
	}
			
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->bpp <= 8)
		{
			UCHAR *imgbuf = (UCHAR *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;
			}
		}
		else if (vinf->bpp <= 16)
		{
			UINT16 *imgbuf = (UINT16 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;				
			}
		}
		else if (vinf->bpp <= 32)
		{
			UINT32 *imgbuf = (UINT32 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;				
			}
		}
		else ERR( -5 );
	}
	
	// drop back buffer, adjust VIDINFHDR
	
	local_framescales = (DOUBLE *) malloc( sizeof(DOUBLE)*1 );
	if (local_framescales == NULL) ERR(-6);
		
	local_framebits = (VOIDP *) malloc( sizeof(VOIDP *)*1 );
	if (local_framebits == NULL) ERR(-6);
	
	local_framebits[0] = (VOIDP) malloc( (vinf->width*vinf->height*vinf->bpp)/8 );
	if (local_framebits[0] == NULL) ERR(-6);


	// free old memory buffers	
	
	if (vinf->framescales != NULL) free( vinf->framescales );
	
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->framebits[num] != NULL) free( (VOIDP) vinf->framebits[num] );
	}
	if (vinf->framebits != NULL) free( (VOIDP) vinf->framebits );
	
	// adjjust vinf header
	vinf->numframes = 1;	
	vinf->framescales = local_framescales;
	vinf->framebits = local_framebits;
	vinf->framescales[0] = vinf->scale;
	
	// update data	
		
	if (vinf->bpp <= 8)
	{
		UCHAR *imgbuf = (UCHAR *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UCHAR) (bufferD[x] / old_num_frames);
		}
	}
	else if (vinf->bpp <= 16)
	{
		UINT16 *imgbuf = (UINT16 *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT16) (bufferD[x] / old_num_frames);
		}
	}
	else if (vinf->bpp <= 32)
	{
		UINT32 *imgbuf = (UINT32 *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT32) (bufferD[x] / old_num_frames);
		}
	}
	
	
err:
	if (errcode != 0)
	{
  		if (local_framescales != NULL) free( (VOIDP) local_framescales );
  		if (local_framebits != NULL)
  		{
  			if (local_framebits[0] != NULL) free( (VOIDP) local_framebits[0] );
  			free( (VOIDP) local_framebits );
  		}
	}
	
	free( (VOIDP) bufferD );

	return errcode;
}


// video_type -> FILE_IMM, FILE_IMC, FILE_BKG, FILE_BKC

// downscale-function: FCN_AVERAGE, FCN_ENVELOPE to create an enveloped or averaged background IF
// the resulting file format will be FILE_BKG (1 image only!)
// FCN_NONE otherwise, if any other type is requested, downscale_function must be FCN_NONE !!

//
// This function is used to conveniently convert between different formats inside a VIDINFHDR. Most conversions
// are lossless, however, conversion of a VIDINFHDR with multiple images to FILE_BKG will be lossy because a
// downscale function is required and will be used.
//
//	Parameters:
//		[in/out]	vinf				VIDINFHDR containing multiple images [in]
//										VIDINFHDR containing the downscaled single image result [out]
//		[in]		newtype				new type inside VIDINFHDR (one of FILE_IMM, FILE_IMC, FILE_BKG, FILE_BKC)
//		[in]		downscale_function	if the resulting format will be FILE_BKG, a downscale function must be passed
//										in order to recalculate the contents. either FCN_AVERAGE or FCN_ENVELOPE might
//										be passed. In any other case one must pass the FCN_NONE statement.
//	Return values:
//
//		 0		okay, everything fine
//		-1		inappropriate 'type' inside vinf structre
//		-2		inappropriate 'type' parameter passed
//		-3		downscale function selected which does not fit to passed structure and resulting type
//		-4		convert to bkg, downscale function was NOT specified
//		-5		unsupported bits per pixel settings inside vinf (using downscaling)
//		-6		memory allocation error (out of memory, using downscale to bkg)
//
// Remarks:
//	In any error case the vinf is not touched and the old data remains inside (without corruption) !
//	FCN_AVERAGE		averaging each frame pixel over the number of frames (average noise rules)
//  FCN_ENVELOPE	the maximum value of each pixel for all frames is obtained and stored as 
//					bitmap (max noise rules)
//
INT32 convert_format( struct VIDINFHDR *vinf, UINT32 newtype, UINT32 downscale_function)
{
	UINT32 ret;
	
	if ((vinf->type != TYPE_IMM) && (vinf->type != TYPE_IMC) && (vinf->type != TYPE_BKC) && (vinf->type != TYPE_BKG))
	{
		return( -1 );
	}

	switch( newtype )
	{
		case TYPE_IMM:
				if (downscale_function != FCN_NONE) return( -3 );
				break;
				
		case TYPE_IMC:
				if (downscale_function != FCN_NONE) return( -3 );
				break;

		case TYPE_BKC:
				if (downscale_function != FCN_NONE) return( -3 );
				break;

		case TYPE_BKG:
				if (vinf->numframes > 1)
				{
					if (downscale_function == FCN_ENVELOPE)
					{
						ret = downscale_envelope_function( vinf );
						if (ret != 0) return( ret );
					}
					else if (downscale_function == FCN_AVERAGE) 
					{
						ret = downscale_average_function( vinf );
						if (ret != 0) return( ret );
					}
					else if (downscale_function == FCN_NONE) return( -4 );
					else return( -3 );
				}
				break;

		default:
				return( -2 );
				break;
	}

	vinf->type = newtype;
	return 0; // all fine
} // convert_format()


//
// This function is the main function to acquire frames from any GrabServer2 program running. It is blocking and
// returns if either one of the error conditions is met or all frames have been acquired successfully.
//
//	Parameters:
//		servername   [in]  hostname of grab server
//		net_port     [in]  port of grab server
//		num_frames   [in]  number of frames to grab as sequence
//		vinf		 [out] empty VIDINFHDR where the acquired frames will be returned
//
//	Return values:
//
//     0		all fine, all frames have been acquired
//    -1		unable to create socket
//    -2		unable to get hostname / malformed ip address
//   -23		vinf structure was not passed empty, please clean up the structure first
//
//		Error emitted by connect() call
//
//    -3		EADDRNOTAVAIL (all in upper case come from connect, see "man connect" for more info)
//    -4		EAFNOSUPPORT
//    -5		EALREADY
//    -6		EBADF
//    -7		ECONNREFUSED
//    -8		EINPROGRESS
//    -9		EINTR
//   -10		EISCONN
//   -11		ENETUNREACH
//   -12		ENOTSOCK
//   -13		EPROTOTYPE
//   -14		ETIMEDOUT
//   -15		EADDRINUSE
//   -16		ECONNRESET
//   -17		EHOSTUNREACH
//   -18		EINVAL
//   -19		ENAMETOOLONG
//   -20		ENETDOWN
//   -21		unknown error at connect statement
//
//		Errors that may happen after a successful connection attempt
//
//   -22		unable to switch to nonblocking mode
//   -24		could not receive video header stream
//   -25		could not receive frame data: timeout
//   -26		could not receive frame data: other errors
//   -27		unable to allocate memory location needed for storing frames
//   -28		frames are not in a sequence 
//	 -29		camera port change has happened during frame receive
//   -30		could not allocate memory for compressed frame bits
//   -31		could not translate vidhdr_stream into VIDHDR
//
//
//	Remarks:
//		Any error after a successful connection attempt alters the vinf structure.
//
INT32 grabframes( const CHAR *servername, const UINT16 net_port, UINT32 num_frames, struct VIDINFHDR *vinf)
{
	struct sockaddr_in	S;
	INT32				errcode=-777;
	socktype			sock = INVALID_SOCKET;
	VIDHDR				lvidH;
  
	if (vinf->type != TYPE_NONE_EMPTY) return -23;

	// create TCP/IP socket
	sock = create_inet_socket(SOCK_STREAM);
  
	if (sock == INVALID_SOCKET) ERR(-1);

	// find out whether the lpszHostAddress is a dotted IP address
	S.sin_family = AF_INET;
	S.sin_port = htons(net_port);
	S.sin_addr.s_addr = inet_addr( servername );

	if (S.sin_addr.s_addr == INADDR_NONE) // aha, no dotted IP address, 
	{                                     // so look up hostname
		// get ip address of server and save it in S (sockaddr_in)
		struct hostent *hentry = gethostbyname(servername);
  
		if (hentry == NULL) ERR(-2);
		S.sin_addr.s_addr = *((UINT32 *)(hentry->h_addr));
	}

	if (connect_socket(sock, (struct sockaddr *) &S, sizeof(S)) == SOCKET_ERROR)
	{
		INT32 ecode = getlasterror();

		close_socket( sock );

		switch( ecode )
		{
			case EADDRNOTAVAIL: ERR( -3 ); break;
			case EAFNOSUPPORT:	ERR( -4 ); break;
			case EALREADY:		ERR( -5 ); break;
			case EBADF:			ERR( -6 ); break;
			case ECONNREFUSED:	ERR( -7 ); break;
			case EINPROGRESS:	ERR( -8 ); break;
			case EINTR:			ERR( -9 ); break;
			case EISCONN:		ERR( -10 ); break;
			case ENETUNREACH:	ERR( -11 ); break;
			case ENOTSOCK:		ERR( -12 ); break;
			case EPROTOTYPE:	ERR( -13 ); break;
			case ETIMEDOUT:		ERR( -14 ); break;
			case EADDRINUSE:	ERR( -15 ); break;
			case ECONNRESET:	ERR( -16 ); break;
			case EHOSTUNREACH:	ERR( -17 ); break;
			case EINVAL:		ERR( -18 ); break;
			case ENAMETOOLONG:	ERR( -19 ); break;
			case ENETDOWN:		ERR( -20 ); break;
			default:			ERR( -21 ); break;
		}
	}

	// now connection to the server is established....
  
	if (set_socket_nonblocking ( sock ) == SOCKET_ERROR) ERR(-22);

	// prepare_vidinf_structure

	{//varblock1
	UINT32	lnumframes=0;
	UINT32	cportcheck=0;
	UINT32	sequencecheck=0;
	UCHAR   vidhdr_stream[REAL_SIZE_VIDHDR];

	vinf->type = TYPE_IMM;

	vinf->framebits = (VOIDP *) malloc( num_frames*sizeof( VOIDP ) );
	memset( vinf->framebits, 0, num_frames*sizeof( VOIDP ) );
	vinf->framescales = (DOUBLE *) malloc( num_frames*sizeof( DOUBLE ) );
	vinf->numframes = num_frames;
	
	while (lnumframes < num_frames ) 
	{
		UINT32 length;
		INT32 ret = receive_socket(sock, vidhdr_stream, REAL_SIZE_VIDHDR, &length, 2000); // read in video header, 
																						   // 2 second timeout
		
		if ((ret != 1) && (length != REAL_SIZE_VIDHDR)) ERR( -24 );
		
		ret = translate_charstream_into_vidhdr( &lvidH, vidhdr_stream);
	  
		if (ret != 0) ERR( -31 );
	  
		// allocate data in order to read in frame data
		// +20 because of safety for decompression
		{//varblock2
		UCHAR *compressed_data = (UCHAR *) malloc( lvidH.len_total - sizeof(VIDHDR) + 20);
		UCHAR *compressed_ptr = (UCHAR *) (compressed_data+lvidH.len_fmthdr);
				
		if (compressed_data == NULL) ERR( -29 );

		ret = receive_socket(sock, (UCHAR *) compressed_data, lvidH.len_total - sizeof(VIDHDR), &length, 2000);
		if (ret == 0 || ret == SOCKET_ERROR || ret == 2)
		{ 
			// error? then out of here
			free( (VOIDP) compressed_data);
			if (ret == 2) ERR( -25 );

			ERR( -26 );
		}

		ret = 2;

		if (lnumframes > 0)
		{
			if (cportcheck != lvidH.cameraport)	ERR( -29 );
			printf("%d\n", lvidH.frameno);
			if (lvidH.frameno != (sequencecheck+1)) ERR( -28 );
			sequencecheck = lvidH.frameno;
		}
		else // initialize checking
		{
			cportcheck = lvidH.cameraport;
			sequencecheck = lvidH.frameno;
		}

		vinf->framebits[lnumframes] = (VOIDP) malloc( (lvidH.widthframe*lvidH.heightframe*lvidH.bpp)/8 );

		if (vinf->framebits[lnumframes] == NULL)
		{
			free( (VOIDP) compressed_data );
			ERR( -27 );
		}
    
		vinf->bpp = lvidH.bpp;
		vinf->bpp_effective = lvidH.bpp_effective;
		vinf->height = lvidH.heightframe;
		vinf->width = lvidH.widthframe;
		vinf->scale = lvidH.scale;
		vinf->framescales[lnumframes] = lvidH.scale;

		{//varblock3
		UINT32 uncompressed_length = (lvidH.widthframe*lvidH.heightframe*lvidH.bpp)/8;
    
		if (lvidH.compressed == 1) // compressed frame ? 
			decompress_huffyuv( vinf->framebits[lnumframes], (const VOIDP) compressed_ptr, uncompressed_length ); // decompress it
		else
			memcpy( vinf->framebits[lnumframes], (const VOIDP) compressed_ptr, uncompressed_length ); // uncompressed: just copy frame
		}//varblock3
		    
		free( (VOIDP) compressed_data ); // delete temporaily allocated space
		}//varblock2
    
		lnumframes++;    
	}
	}//varblock1
  
	
	close_socket( sock );
	sock = INVALID_SOCKET;
  
	return( 0 );

err:
	if (sock != INVALID_SOCKET)	close_socket( sock );
	if (vinf->type == TYPE_IMM) 
	{
		free_vidinfhdr( vinf );
	}

	return( errcode );
}
