// BMPFileHelper.cpp: implementation of the CBMPFileHelper class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "videoclnt.h"
#include "headers.h"
#include "MemoryManagement.h"
#include "BMPFileHelper.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction (just empty)
//////////////////////////////////////////////////////////////////////

CBMPFileHelper::CBMPFileHelper()
{

}

CBMPFileHelper::~CBMPFileHelper()
{

}

///////////////////////////////////////////////////////////////////////////////
//
// This helper function writes out a single bmp file (8 bits per pixel using 
// color palette). It returns TRUE
// if the file was successfully written, otherwise it returns FALSE.
//
// Parameters:
// szFile    - filename and path
// imagebits - memory location where the actual bitmap bits are stored
// pal       - Palette for the BMP file
// adjusted  - TRUE: image is aligned to LONG boundary
//             FALSE: image is not yet aligned to LONG boundary
//
BOOL CBMPFileHelper::WriteBMP( LPTSTR szFile, void *imagebits, PALETTEENTRYSW *hpal, int Width, int Height, int physbpp, int effbpp, BOOL adjusted)
{
	BITMAPINFO *BMInfo;
	BITMAPFILEHEADER FileHeader;

	int addWidth = Width+Width%4;

	// allocate sufficent bitmapinfo structure
	BMInfo = (BITMAPINFO *) memman.sw_malloc( sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD));
	if (BMInfo == NULL) return FALSE;
 
	// calculate file size (neede for header)
	long FileSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) +
		     256 * sizeof(RGBQUAD) + addWidth * Height;

	// fill in FileHeader

	FileHeader.bfType = 0x4D42;
	FileHeader.bfSize = FileSize;  // should be specified in bytes
	FileHeader.bfReserved1 = 0;
	FileHeader.bfReserved2 = 0;
	FileHeader.bfOffBits = sizeof(BITMAPINFOHEADER) + 256 
	* sizeof(RGBQUAD) + sizeof(BITMAPFILEHEADER);  // distance to bitmap data

	// fill in InfoHeader

	BMInfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	BMInfo->bmiHeader.biWidth = Width;  
	BMInfo->bmiHeader.biHeight = Height; 
	BMInfo->bmiHeader.biPlanes = 1;  // should be 1
	BMInfo->bmiHeader.biBitCount = 8; // should be 8
	BMInfo->bmiHeader.biCompression = BI_RGB;
	BMInfo->bmiHeader.biSizeImage = addWidth * Height;
	BMInfo->bmiHeader.biXPelsPerMeter = 0;
	BMInfo->bmiHeader.biYPelsPerMeter = 0;
	BMInfo->bmiHeader.biClrUsed = 256;
	BMInfo->bmiHeader.biClrImportant = 256;



	// fill in palette entries
	for (unsigned int i=0;i<256;i++)
	{
		unsigned int indexmax = (1 << effbpp)-1;
		unsigned int index = (unsigned int) ((indexmax/255.0)*i);
		if (index > indexmax) index = indexmax; // for safety

		PALETTEENTRYSW *loc = &hpal[index];

		BMInfo->bmiColors[i].rgbRed = loc->peRed;
		BMInfo->bmiColors[i].rgbGreen = loc->peGreen;
		BMInfo->bmiColors[i].rgbBlue = loc->peBlue;
		BMInfo->bmiColors[i].rgbReserved = 0;
	}

	//unsigned long SizeFileBmp = addWidth * Height;

	// open file
	CFile file;
	if( !file.Open( szFile, CFile::modeWrite|CFile::modeCreate) )
		return FALSE; // return error if file could not be opened


	// allocate temp buffer for image conversion
	unsigned char *tmpbuffer = (unsigned char *) memman.sw_malloc( addWidth*Height );
	if (tmpbuffer == NULL) 
	{
		file.Close();
		memman.sw_free( (void *) BMInfo );
		return FALSE;
	}

	unsigned char *ready_image = (unsigned char *) imagebits;
	unsigned char *downscaled_image = NULL;

	if (physbpp > 8) 
	{
		downscaled_image = (unsigned char *) memman.sw_malloc( Width * Height );
		if (downscaled_image == NULL) 
		{
			file.Close();
			memman.sw_free( (void *) BMInfo );
			memman.sw_free( (void *) tmpbuffer );
			return FALSE;
		}
		
		DownScaleTo8Bit( (unsigned char *) imagebits, downscaled_image, Width, Height, physbpp);
		ready_image = downscaled_image;
		adjusted = FALSE;
	}

	// flip frame from top-down to bottom-up ( images are flipped 
	// around the horizontal axis ina BMP file )
	if (adjusted == TRUE) flipframemmx_without_adjustment( tmpbuffer, ready_image, addWidth, Height, 8 ); //1
                         else flipframemmx_with_adjustment( tmpbuffer, ready_image, Width, Height, 8 );

	// write out fileheader
	file.Write( &FileHeader, sizeof(FileHeader) );

	// write out bitmapinfo structure 
	file.Write( (const void *) BMInfo, sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD) );

	// Write the bits 
	file.Write( (const void *) tmpbuffer, addWidth*Height );

	// close file
	file.Close();

	// free temporary allocated memory
	memman.sw_free( (void *) tmpbuffer );

	if (downscaled_image != NULL) memman.sw_free( (void *) downscaled_image );

	// free memory needed for bitmapinfo structure
	memman.sw_free( (void *) BMInfo );

	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// This helper function loads in a single bmp file (8 bits per pixel using 
// color palette, not compressed). It returns TRUE
// if the file was successfully loaded, otherwise it returns FALSE.
// This function was taken from a newsgroup article and is not written by the 
// author of this program.
//
// Parameters:
//  lpszFileName    - filename and path
//  ppvbits         - pointer to where a pointer is stored after calling this function
//                    using this pointer you are able to access the bits, can be NULL
//  hSection        - not used here, can be NULL
//  dwOffset        - not used here, should be set to 0 
//
// Return value:
//  NULL            - could not load file 
//  otherwise       - bitmap handle to access the loaded image
//
HBITMAP CBMPFileHelper::LoadDIBSectionFromFile( LPCTSTR lpszFileName, LPVOID *ppvBits, 
					HANDLE hSection, DWORD dwOffset) 
{
	LPVOID lpBits;
	CFile file;

    // open file 
	if( !file.Open( lpszFileName, CFile::modeRead) )
		return NULL; // file could not be opened, return NULL
	
	BITMAPFILEHEADER bmfHeader;
	long nFileLen;
	
	nFileLen = file.GetLength();
	
	// Read file header
	if (file.Read((LPSTR)&bmfHeader, sizeof(bmfHeader)) != sizeof(bmfHeader))
		return NULL;
	
	// File type should be 'BM'
	if (bmfHeader.bfType != ((WORD) ('M' << 8) | 'B'))
		return NULL;

    // allocate bitmapinfo structure
	BITMAPINFO *pbmInfo;
	pbmInfo = (BITMAPINFO *)::GlobalAlloc(GMEM_FIXED, 
				sizeof(BITMAPINFO) + sizeof(RGBQUAD)*256 );
	if (pbmInfo == NULL)
		return NULL;
	
	// Read the BITMAPINFO
	file.Read( pbmInfo, sizeof(BITMAPINFO) + sizeof(RGBQUAD)*256 );
	
	BITMAPINFO &bmInfo = *pbmInfo;
	
	if (bmInfo.bmiHeader.biCompression != BI_RGB)
	{
     // free temporary allocated memory
	 ::GlobalFree(pbmInfo);
	 return NULL;
    }

	// create DIB section for image
	HBITMAP hBmp = CreateDIBSection( NULL, pbmInfo, DIB_RGB_COLORS, &lpBits, 
						hSection, dwOffset );

    if (hBmp != NULL)
	{
 	  LPBYTE  lpDIBBits;              // Pointer to DIB bits
	  int nColors = bmInfo.bmiHeader.biClrUsed ? bmInfo.bmiHeader.biClrUsed : 
				1 << bmInfo.bmiHeader.biBitCount;
	
	  if( bmInfo.bmiHeader.biBitCount > 8 )
		lpDIBBits = (LPBYTE)((LPDWORD)(bmInfo.bmiColors + 
			bmInfo.bmiHeader.biClrUsed) + 
			((bmInfo.bmiHeader.biCompression == BI_BITFIELDS) ? 3 : 0));
	  else
		lpDIBBits = (LPBYTE)(bmInfo.bmiColors + nColors);

	  int nOffset = sizeof(BITMAPFILEHEADER) + (lpDIBBits - (LPBYTE)pbmInfo);
	  file.Seek( nOffset, CFile::begin);
	
	  // read in file bits
	  file.ReadHuge((LPSTR)lpBits, nFileLen - nOffset); //bmInfo.biSizeImage ); 

      // free temporary allocated memory
	  ::GlobalFree(pbmInfo);

      // if pointer to pointer is not NULL store the image bits
	  if( ppvBits )
		  *ppvBits = lpBits;
	}

    file.Abort();

	return hBmp;
}

////////////////////////////////////////////////////////////////////////
// DownScaleTo8Bit() 
//
// This mathod scales image src (attributes: width, height and bpp) into 
// 8 bits memory location dst (attributes: width, height, bpp=8).
//
// One has to take care that memory location src is at least as big as 
//  width*height*bpp/8 bytes
// and that dst is at least as big as
//  width*height bytes
//

void CBMPFileHelper::DownScaleTo8Bit(unsigned char *src, unsigned char *dst, unsigned int width, unsigned int height, unsigned int bpp)
{
	if (bpp == 8)  // 8 bits per pixel, then just copy the data
	{
		memman.memcpysw( (char *) dst, (char *) src, width*height);
	}
	else if (bpp == 16) // 16 bits per pixel 
	{
		unsigned short maxval=0;
		unsigned short *srci = (unsigned short *) src;
 
		// look into source for maximum value
		for (unsigned int j=0;j<width*height;j++)
		{
			if (srci[j] > maxval) maxval = srci[j];
		}
   
		// scale into 8 bits accordingly to the maximum value
		unsigned long tmp = 65536L / (((unsigned long)maxval)+1);
		unsigned short tmp1 = (unsigned short) tmp;
		for (j=0;j<width*height;j++)
		{
			dst[j] = (unsigned char) ((srci[j] * tmp1) >> 8);
		}
	}
	else if (bpp == 32) // 32 bits per pixel
	{
		unsigned long *srcl = (unsigned long *) src;
		unsigned long maxval=0;
 
		// look into source for maximum value
		for (unsigned int j=0;j<width*height;j++)
		{
			if (srcl[j] > maxval) maxval = srcl[j];
		}
   
		// scale into 8 bits accordingly to the maximum value
		unsigned long tmp = 0xFFFFFFFF / (maxval);
		for (j=0;j<width*height;j++)
		{
			dst[j] = (unsigned char) ((srcl[j] * tmp) >> 24);
		}
	}
 
 }



////////////////////////////////////////////////////////////////////////
// flipframemmx - flips a video frame horizontally and alignes the destination
//				  scanlines to LONG boundary
//
// This function flips a video frame horizontally. The first line of 'src' becomes
// the last line of 'dst', the last line of 'src' becomes the first line of 'dst' etc.
// 
// Only 8 bits per pixel are tested, other bpp are not tested
//
void CBMPFileHelper::flipframemmx_with_adjustment(unsigned char *dest, unsigned char *src, int width, int height, int bpp)
{
  unsigned char *srcptr = src;
  unsigned char *destptr = dest;
  int addWidth = width+width%4;
  int lastlinesrc = (width*height*bpp/8)-(width*bpp/8);

  srcptr += lastlinesrc;
  int subsrc = width*bpp/8;
  int adddst = addWidth*bpp/8;

  for (int i=0;i<height;i++)
  {
   memcpy( (void *) destptr, (const void *) srcptr, width);
   destptr += adddst;
   srcptr -= subsrc; 
  }
}

////////////////////////////////////////////////////////////////////////
// flipframemmx - flips a video frame horizontally without aligning the destination
//				  scanlines to LONG boundary
//
// This function flips a video frame horizontally. The first line of 'src' becomes
// the last line of 'dst', the last line of 'src' becomes the first line of 'dst' etc.
// 
// Only 8 bits per pixel are tested, other bpp are not tested
//
void CBMPFileHelper::flipframemmx_without_adjustment(unsigned char *dest, unsigned char *src, int width, int height, int bpp)
{
    unsigned char *srcptr = src;
    unsigned char *destptr = dest;
    int lastline = (width*height*bpp/8)-(width*bpp/8);

    srcptr += lastline;

    int addsrc = (width*bpp/8);
    int adddst = addsrc;

    for (int i=0;i<height;i++)
    {
     memcpy( (void *) destptr, (const void *) srcptr, width);
     destptr += adddst;
     srcptr -= addsrc; 
    }
}
