// EnvelopeAverage.cpp: implementation of the CEnvelopeAverage class.
//
//////////////////////////////////////////////////////////////////////
// CEnvelopeAverage.cpp
//
// This is a helper class for Envelope and Average background creation.
// One just Init() the class, AddFrame() frame after frame, return 
// the averaged or enveloped background using ReturnBuffer() and de-
// initializes the class using Done().
//
//
//////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "videoclnt.h"
#include "headers.h"
#include "MemoryManagement.h"
#include "EnvelopeAverage.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// construction, just initialize the class to not_initialized state
CEnvelopeAverage::CEnvelopeAverage()
{
 m_bSuccessfullyInitialized = FALSE;
}

// destruction, frees all buffers if class is still initialized
CEnvelopeAverage::~CEnvelopeAverage()
{
 if (m_bSuccessfullyInitialized == TRUE)
 {
	 Done();
 }
}

//////////////////////////////////////////////////////////////////////////////////////
// CEnvelopeAverage::Init( unsigned int bpp, VIDHDR vidH, BOOL envelope, BOOL xray )
//
// This function initislizes the class in order to create an envelope or 
// average background.
//
// Parameters:
//  bpp       - bits per pixel setting, values of 8, 16 and 32 are supported
//  vidH      - example video header of one frame that will be processed
//  emvelope  - TRUE  : create envelope background
//              FALSE : create average background
//  xray      - TRUE  : filter the xrays out before processing a frame
//              FALSE : do not filter the xrays 
//
BOOL CEnvelopeAverage::Init( unsigned int bpp, VIDHDR vidH, BOOL envelope, BOOL xray )
{
 // if already initialized -> return FALSE
 if (m_bSuccessfullyInitialized == TRUE) return FALSE;

 // store important settings into member variables
 m_bEnvelope = envelope;
 m_iBpp = bpp;
 m_iWidth = vidH.widthframe;
 m_iHeight = vidH.heightframe;
 m_iCount = 0;
 m_bXRay = xray;

 if ((bpp == 8) || (bpp == 16)) // 8 or 16 bits per pixel ? 
 {
  // create integer buffer 
  m_pBuffer816 = (unsigned int *) memman.sw_malloc( vidH.widthframe*vidH.heightframe*sizeof(unsigned int) );
  if (m_pBuffer816 == NULL) return FALSE; // could not allocate -> return FALSE

  // reset buffer to zeroes
  memset( (void *) m_pBuffer816, 0, vidH.widthframe*vidH.heightframe*sizeof(unsigned int));

  // we are properly initialized
  m_bSuccessfullyInitialized = TRUE;
  // success
  return TRUE;
 }
 else
 {
  // create double buffer 
  m_pBuffer32 = (double *) memman.sw_malloc( vidH.widthframe*vidH.heightframe*sizeof(double) );
  if (m_pBuffer32 == NULL) return FALSE; // could not allocate -> return FALSE

  // reset buffer to zeroes
  for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
  {
	m_pBuffer32[i] = 0.0;
  }

  // we are properly initialized
  m_bSuccessfullyInitialized = TRUE;
  // success
  return TRUE;
 }

 // failure
 return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////////
// CEnvelopeAverage::AddFrame( unsigned char *frame, VIDHDR vidH)
//
// This function adds a single frame to the envelope/average background 
//
// Parameters:
//  frame     - pure frame bits of the frame
//              must contain at least vidH.widthframe*vidH.heightframe*vidH.bpp/8 bytes
//  vidH      - video header of the frame bits
//
BOOL CEnvelopeAverage::AddFrame( unsigned char *frame, VIDHDR vidH)
{
 // initialize default variables
 unsigned char *xraybuffer = NULL;
 unsigned char *lframe = frame;

 // use xray-filtering ?
 if (m_bXRay == TRUE)
 {
  // create xray buffer
  xraybuffer = (unsigned char *) memman.sw_malloc( vidH.widthframe*vidH.heightframe*vidH.bpp/8 );
  if (xraybuffer == NULL) return FALSE; // could not allocate -> return error

  // copy frame bits into xray buffer
  memman.memcpysw( (char *) xraybuffer, (char *) frame, vidH.widthframe*vidH.heightframe*vidH.bpp/8 );

  lframe = xraybuffer;

  // filter the xrays 
  XFilter( xraybuffer, m_iWidth, m_iHeight);
 }

 if (m_bEnvelope == TRUE) // envelope ? 
 {
  if (m_iBpp == 8) // 8 bits per pixel
  {
	  unsigned char *frm = lframe;
      // loop over the whole frame array
	  for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
	      unsigned int n = (unsigned int) frm[i];
	      if (m_pBuffer816[i] < n ) m_pBuffer816[i] = n;
	  }
  }
  else if (m_iBpp == 16) // 16 bits per pixel
  {
	  unsigned short *frm = (unsigned short *) lframe;
	  // loop over the whole frame array
      for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
		  unsigned int n = (unsigned int) frm[i];
	      if (m_pBuffer816[i] < n ) m_pBuffer816[i] = n;
	  }
  }
  else // 32 bits per pixel
  {
	  unsigned long *frm = (unsigned long *) lframe;
	  // loop over the whole frame array
      for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
		  double n = (double) frm[i];
	      if (m_pBuffer32[i] < n ) m_pBuffer32[i] = n; 
	  }
  }
 }
 else // AVERAGE
 {
  if (m_iBpp == 8) // 8 bits per pixel
  {
	  unsigned char *frm = lframe;
	  // loop over the whole frame array
      for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
	      unsigned int n = (unsigned int) frm[i];
	      m_pBuffer816[i] += n; 
	  }
  }
  else if (m_iBpp == 16) // 16 bits per pixel
  {
	  unsigned short *frm = (unsigned short *) lframe;
	  // loop over the whole frame array
      for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
		  unsigned int n = (unsigned int) frm[i];
	      m_pBuffer816[i] += n; 
	  }
  }
  else // 32 bits per pixel
  {
	  unsigned long *frm = (unsigned long *) lframe;
	  // loop over the whole frame array
      for (unsigned int i=0;i<vidH.widthframe*vidH.heightframe;i++)
	  {
		  double n = (double) frm[i];
	      m_pBuffer32[i] += n; 
	  }
  }
 }

 // xray buffer was allocated? -> free it
 if (xraybuffer != NULL) memman.sw_free( (void *) xraybuffer );

 // increase count by 1 (picture)
 m_iCount++;
 return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////////
// CEnvelopeAverage::ReturnBuffer( unsigned char *outbuffer, unsigned int *size )
//
// This function returns the background buffer to be used as envelope or average 
// background
//
// Parameters:
//  outbuffer - memory location where to store the calculated background bits
//              NULL: required size of the output buffer in bytes will be put in size
//  size      - [out] 
//              out: (outbuffer==NULL) returns size needed for output buffer
//
void CEnvelopeAverage::ReturnBuffer( unsigned char *outbuffer, unsigned int *size )
{
 if (size != NULL) *size = m_iWidth*m_iHeight*m_iBpp/8; // return size needed

 if (outbuffer == NULL) return;

 if (m_bEnvelope == TRUE) // envelope ? 
 {
  if (m_iBpp == 8) // 8 bits per pixel ? 
  {
	  unsigned char *outbuf = outbuffer;
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned char) m_pBuffer816[i];
      }
  }
  else if (m_iBpp == 16) // 16 bits per pixel ? 
  {
	  unsigned short *outbuf = (unsigned short *) outbuffer;
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned short) m_pBuffer816[i];
      }
  }
  else // 32 bits per pixel  
  {
	  unsigned long *outbuf = (unsigned long *) outbuffer;
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned long) m_pBuffer32[i];
      }
  }
 }
 else
 {
  if (m_iBpp == 8) // 8 bits per pixel ? 
  {
	  unsigned char *outbuf = outbuffer;
	  // put background buffer to outbuffer 
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned char) (m_pBuffer816[i] / m_iCount );
      }
  }
  else if (m_iBpp == 16) // 16 bits per pixel ? 
  {
	  unsigned short *outbuf = (unsigned short *) outbuffer;
	  // put background buffer to outbuffer 
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned short) (m_pBuffer816[i] / m_iCount );
      }
  }
  else // 32 bits per pixel
  {
	  unsigned long *outbuf = (unsigned long *) outbuffer;
	  // put background buffer to outbuffer 
	  for (unsigned int i=0;i<m_iWidth*m_iHeight;i++)
	  {
		  outbuf[i] = (unsigned long) (m_pBuffer32[i] / m_iCount );
      }
  }
 }

}

//////////////////////////////////////////////////////////////////////////////////////
// CEnvelopeAverage::Done()
//
// This function deinitializes the class and frees the buffers.
//
void CEnvelopeAverage::Done()
{
 // are we initialized ? 
 if (m_bSuccessfullyInitialized == TRUE)
 {
	 if (m_iBpp == 32) // 32 bits per pixel ? 
	 {
      memman.sw_free( (void *) m_pBuffer32 ); // free big buffer
	 }
	 else // 8 or 16 bpp ? 
	 {
      memman.sw_free( (void *) m_pBuffer816 ); // free small buffer
	 }
     // we are not initialized any more
     m_bSuccessfullyInitialized = FALSE;
 }
 
}

//////////////////////////////////////////////////////////////////////////////////////
// CEnvelopeAverage::XFilter()
//
// This function filters the XRay's out of 'data'. Width and height give the width
// and height of the image in pixels. The member variable m_iBpp is used to translate
// the pixels of width and height to bytes.
//
// Parameters:
//  data       - data buffer where the xrayed frame bits are stored
//  width      - width of the image in pixels 
//  height     - height of the image in pixels 
//             
//
void CEnvelopeAverage::XFilter(unsigned char *data, int width, int height)
{
  int up = 0;
  int left = 0;
  int down = up+height-1;
  int right = left+width-1;
  int ImageWidth = m_iWidth;

  if (m_iBpp == 8)
  {
    unsigned char *data8 = (unsigned char *)data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		int equalize=0;
		const int datapos = data8[pos];
		if(datapos)
		{
			const int datapos2 = (datapos * 2) / 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=((data8[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth+1])<datapos2);


			if(equalize>3)
  /*equalize*/{
				unsigned int sum=0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum >>= 3;
				
				data8[pos]=(unsigned char)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }
  else if (m_iBpp == 16)
  {
    unsigned short *data16 = (unsigned short *) data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		int equalize=0;
		const int datapos = data16[pos];
		if(datapos)
		{
			const unsigned int datapos2 = (datapos * 2) / 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=((data16[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth+1])<datapos2);


			if(equalize>3)
  /*equalize*/{
				unsigned int sum=0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum >>= 3;
				
				data16[pos]=(unsigned short)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }
  else // bpp == 32
  {
    unsigned int *data32 = (unsigned int *) data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		double equalize=0;
		const unsigned int datapos = data32[pos];
		if(datapos)
		{
			const unsigned int datapos2 = (datapos / 2) * 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=(double) ((data32[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth+1])<datapos2);


			if(equalize>3.0)
  /*equalize*/{
				double sum=0.0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum /= 8.0;
				
				data32[pos]=(unsigned int)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }

}
