// CFourierStat.cpp: implementation of the CCFourierStat class.
//
// Initially written by Velizar Miltchev (DESY Zeuthen) miltchev@ifh.de
// some additional code by Stefan Weisse sweisse@ifh.de
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "headers.h"
#include "videoclnt.h"
#include "FourierStat.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// 
// The constructor of CFourierStat does some basic initialization. In addition, the background image
// is subtracted, the area of interest is normalized and the image is (maybe) x-ray filtered.
// The projections are calculated, too. Please note that due to heavy optimization the ImWi and ImHi
// parameter must be have their default value. (ImWi=768 ImHi=574)
//
// Parameters:
//  data     - pointer to the to be analysed images (must contain at least (ImWi*ImHi bytes)
//  back     - pointer to the to be subtracted background image (must contain at least ImWi*ImHi bytes)
//  out      - pointer where the analyzed image should be stored (must contain at least ImWi*ImHi bytes)
//  AOI      - rectangular region of area of interest inside frame - ranges must be x = [0,ImWi] y = [0,ImHi]
//  XFilt    - x-ray filtering TRUE=on FALSE=off
//  deg      - degree
//  recdep   - recursion depth of fourier analysis
//  ImWi     - width of the image: _must_ be set to 768 (because of tight optimization)
//  ImHi     - height of the image: _must_ be set to 574 (because of tight optimization)
//
CFourierStat::CFourierStat(unsigned char* data,unsigned char* Back, unsigned char *out, CRect AOI,BOOL XFilt, BOOL circular, BOOL normalize, 
		unsigned short deg,unsigned short recdep,unsigned int ImWi,unsigned int ImHi, unsigned int bpp, unsigned int bpp_eff)
		:CStat(data,Back,out,AOI,XFilt,circular, normalize, ImWi,ImHi, bpp, bpp_eff) // most initialization is handled by the CStat construction
{
	// initialize degree and RecursDepth members
	degree=deg;
	RecursDepth=recdep;
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// destruction of the class: empty
//
CFourierStat::~CFourierStat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the A coefficent needed by fourier analysis. The function does
// not use the member variables of the class, but the parameters.
//
// Parameters:
//  DistArray - distribution array to analyse
//  len       - length of the distribution array
//  deg       - number of degrees
//
// Return value:
//  Coefficent array that was dynamically allocated
//
double* CFourierStat::CalcAcoef(double* DistArray,int len, unsigned short deg)
{
	double locl=len;
	double wavenum=2*pi/(locl+1e-10);
	double* coefarr= (double *) memman.sw_malloc( sizeof(double)*degree );
	if (coefarr == NULL) return NULL;
	ZeroMemory(coefarr,degree*sizeof(double));

	double constant = 2.0 / (1e-10+(double) len );
	
	for( short i=0;i<degree;i++)
	{
	    double start = i*wavenum;
		double akt = 0;
		for(int j=0;j<=len;j++)
		{
			coefarr[i]+=constant*DistArray[j]*cos( akt );
	        akt += start;
		}
	}
	return coefarr;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the B coefficent needed by fourier analysis. The function does
// not use the member variables of the class, but the parameters.
//
// Parameters:
//  DistArray - distribution array to analyse
//  len       - length of the distribution array
//  deg       - number of degrees
//
// Return value:
//  Coefficent array that was dynamically allocated. Make sure to deallocate it.
//
double* CFourierStat::CalcBcoef(double* DistArray,int len, unsigned short deg)
{	

	double locl=len;
	double wavenum=2*pi/(locl+1e-10);
	double* coefarr= (double *) memman.sw_malloc( sizeof(double)*degree );
	if (coefarr == NULL) return NULL;
	ZeroMemory(coefarr,degree*sizeof(double));
	
	double constant = 2.0 / (1e-10+(double) len );

	for( int i=0;i<degree;i++)
	{
	    double start = i*wavenum;
		double akt = 0;
		for(int j=0;j<=len;j++)
		{
			coefarr[i]+=constant*DistArray[j]*sin( akt );
			akt += start;
		}
	}
	return coefarr;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function recalculates the XDIST distribution array out of the base data found in XDIST
// distribution array. The recalculation is needed because the fourier analysis needs a 
// different distribution array than calculations done in CStat.
//
// Parameters:
//  (XDIST)  - no real parameter, but member variable which needs to be properly initialized
//  locleft  - starting index inside XDIST array 
//  locright - ending index inside XDIST array 
//
// Return value:
//  ReXDist  - calculated Fourier distribution array. This array is dynamically allocated so 
//             make sure that it gets properly deallocated.
//
double* CFourierStat::ReXDist(int locleft, int locright)
{
    // calculate len based on indices locleft and locright
	double len=((double)(locright-locleft));len+=1e-10;
	
	// allocate 'Re' distribution array
	double* ReDist= (double *) memman.sw_malloc( sizeof(double)*ImageWidth );
	
	// calculate A and B coefficents
	double* coefA=CalcAcoef(&XDIST[locleft],(int)len,degree);
	double* coefB=CalcBcoef(&XDIST[locleft],(int)len,degree);

	if (coefA == NULL || coefB == NULL) 
	{
	 if (coefA != NULL) memman.sw_free( (void *) coefA );
	 if (coefB != NULL) memman.sw_free( (void *) coefB );
	 return NULL;	
	}

    // calculate distribution array based on coefficents
	double wavenum=2*pi/len;
	double sum=1e-10;
	for(int i=locleft;i<=locright;i++)
	{
		ReDist[i]=coefA[0]/2;
		double start = ((double) (i-locleft))*wavenum;
		double akt = start;

		for(int j=1;j<degree;j++)
		{
		  ReDist[i]+=coefA[j]*cos( akt )+ coefB[j]*sin(akt);
		  akt += start;
        }
		if (ReDist[i] < 0) ReDist[i] = 0;
		//ReDist[i]*=(ReDist[i]>0);
		sum+=(ReDist[i]+=1e-10);
	}
	
	// normalize distribution array
	for(i=locleft;i<=locright;i++)ReDist[i]/=sum;
	
	// free up memory
	memman.sw_free( (void *) coefA);
	memman.sw_free( (void *) coefB);
	
	// return distribution array
	return ReDist;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function recalculates the YDIST distribution array out of the base data found in YDIST
// distribution array. The recalculation is needed because the fourier analysis needs a 
// different distribution array than calculations done in CStat.
//
// Parameters:
//  (YDIST)  - no real parameter, but member variable which needs to be properly initialized
//  locup    - starting index inside YDIST array 
//  locdown  - ending index inside YDIST array 
//
// Return value:
//  ReYDist  - calculated Fourier distribution array. This array is dynamically allocated so 
//             make sure that it gets properly deallocated.
//
double* CFourierStat::ReYDist(int locup, int locdown)
{
	// calulate len based on indices locup and locdown
	double len=((double)(locdown-locup));len+=+1e-10;
	
	// allocate 'Re' distribution array
	double* ReDist= (double *) memman.sw_malloc( sizeof(double)*ImageHeight );
	
	// calculate A and B coefficents
	double* coefA=CalcAcoef(&YDIST[locup],(int)len,degree);
	double* coefB=CalcBcoef(&YDIST[locup],(int)len,degree);

	if (coefA == NULL || coefB == NULL) 
	{
	 if (coefA != NULL) memman.sw_free( (void *) coefA );
	 if (coefB != NULL) memman.sw_free( (void *) coefB );
	 return NULL;	
	}
	
	// calculate distribution array based on coefficents
	double wavenum=2*pi/len;
	double sum=1e-10;
	for(int i=locup;i<=locdown;i++)
	{
		ReDist[i]=coefA[0]/2;
		double start = ((double) (i-locup))*wavenum;
		double akt = start;

		for(int j=1;j<degree;j++)
		{
		  ReDist[i]+=coefA[j]*cos( akt )+ coefB[j]*sin(akt);
		  akt += start;
        }
		if (ReDist[i] < 0) ReDist[i] = 0;
		//ReDist[i]*=(ReDist[i]>0);
		sum+=(ReDist[i]+=1e-10);
 /*
		ReDist[i]=coefA[0]/2;
		for(int j=1;j<degree;j++)ReDist[i]+=coefA[j]*cos(((double)(j*(i-locup)))*wavenum)+
			                                coefB[j]*sin(((double)(j*(i-locup)))*wavenum);
		ReDist[i]*=(ReDist[i]>0);
		sum+=(ReDist[i]+=1e-10);
*/
	}
	
	// normalize ReDist array
	for(i=locup;i<=locdown;i++) ReDist[i]/=sum;
	
	// free dynamically allocated A and B coefficents
	memman.sw_free( (void *) coefA);
	memman.sw_free( (void *) coefB);
	
	// return recalculated distribution
	return ReDist;
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the XMEAN (spot center in x dimension) variable out of the passed
// parameters. The function calls itself recursively up to a depth of 'recnum'.
//
// Parameters:
//  dist     - distribution array to calculate x mean value
//  recnum   - recursion depth
//  locleft  - starting index inside dist array 
//  locright - ending index inside dist array 
//
// Return value:
//  xmean    - calculated x mean value (spot center position in x dimension)
//
double CFourierStat::CalcXMEAN( double* dist,unsigned short recnum, int locleft, int locright)
{
 double localxm=locleft,localxs=0;
 for(int i=locleft;i<=locright;i++)localxm=dist[i]>dist[(int)localxm]?i:localxm;
 if(recnum==1)return localxm;
 double sum=1e-10;
 for(i=locleft;i<=locright;i++)sum+=dist[i];
 for(i=locleft;i<=locright;i++)localxs+=(i-localxm)*(i-localxm)*dist[i]/sum;
 localxs=sqrt(localxs);
 int newleft=(int) (localxm-localxs>locleft?localxm-localxs:locleft);
 int newright=(int) (localxm+localxs<locright?localxm+localxs:locright);
 if(newright-newleft<2){newleft=locleft;newright=locright;}
 return CalcXMEAN(dist,recnum-1,newleft,newright);
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the YMEAN (spot center in y dimension) variable out of the passed
// parameters. The function calls itself recursively up to a depth of 'recnum'.
//
// Parameters:
//  dist     - distribution array to calculate y mean value
//  recnum   - recursion depth
//  locleft  - starting index inside dist array 
//  locright - ending index inside dist array 
//
// Return value:
//  ymean    - calculated y mean value (spot center position in y dimension)
//
double CFourierStat::CalcYMEAN( double* dist,unsigned short recnum, int locleft, int locright)
{
 double localxm=locleft,localxs=0;
 for(int i=locleft;i<=locright;i++)localxm=dist[i]>dist[(int)localxm]?i:localxm;
 if(recnum==1)return localxm;
 double sum=1e-10;
 for(i=locleft;i<=locright;i++)sum+=dist[i];
 for( i=locleft;i<=locright;i++)localxs+=(i-localxm)*(i-localxm)*dist[i]/sum;
 localxs=sqrt(localxs);
 int newleft=(int)(localxm-localxs>locleft?localxm-localxs:locleft);
 int newright=(int)(localxm+localxs<locright?localxm+localxs:locright);
 if(newright-newleft<2){newleft=locleft;newright=locright;}
 return CalcYMEAN(dist,recnum-1,newleft,newright);
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function does all calculations. It should be called right after the construction of
// the class and before calling any of the public member functions.
//
void CFourierStat::START()
{
    // calculate fourier distribution arrays
	FTXdist=ReXDist(left,right);
	FTYdist=ReYDist(up,down);

	if (FTXdist != NULL && FTYdist != NULL)
	{
	 // calc mean values (spot center coordinates)
	 XMEAN=CalcXMEAN(FTXdist,RecursDepth,left,right);
	 YMEAN=CalcYMEAN(FTYdist,RecursDepth,up,down); 
	
	 // calculates size values (spot size in x and y dimension)
	 XRMS=CalcXRMS(FTXdist,XMEAN);
	 YRMS=CalcYRMS(FTYdist,YMEAN);
	}
	
	// free up memory allocated while calculating fourier distribution arrays
	if (FTXdist != NULL) memman.sw_free( (void *) FTXdist );
	if (FTYdist != NULL) memman.sw_free( (void *) FTYdist );
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the spot size in x dimension. It does not use any member variables
// but the parameters instead.
//
// Parameters:
//  dist  - distribution array on which calculation is based
//  xm    - previously calculated x mean value
//
// Return value:
//  x rms - spot size in x coordinates
//
double CFourierStat::CalcXRMS(double *dist, double xm)
{
	int lleft = left;
	int lright = right;

	double retv=0;
	double sum=1e-10;
	int lxr= (int) (2.78*CStat::CalcXRMS(XDIST,xm));
	int ll=(int)(xm-lxr>lleft?xm-lxr:lleft);
	int rl=(int)(xm+lxr<lright?xm+lxr:lright);
	lleft=(unsigned int)xm;lright=(unsigned int)xm;
	for(int i=(int)xm;i>=ll;i--)lleft=(int)(dist[i]<dist[lleft]?i:lleft);
	for(i=(int)xm;i<=rl;i++)lright=dist[i]<dist[lright]?i:lright;
	for(i=left;i<=(int)lright;i++)sum+=XDIST[i];
	for(i=left;i<=(int)lright;i++)retv+=(i-xm)*(i-xm)*XDIST[i]/sum;
	return sqrt(retv);
}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// This function calculates the spot size in y dimension. It does not use any member variables
// but the parameters instead.
//
// Parameters:
//  dist  - distribution array on which calculation is based
//  ym    - previously calculated y mean value
//
// Return value:
//  y rms - spot size in y coordinates
//
double CFourierStat::CalcYRMS(double *dist, double ym)
{
	int lup = up;
	int ldown = down;

	double retv=0;
	double sum=1e-10;
	int lxr=(int)(2.78*CStat::CalcYRMS(YDIST,ym));
	int ll=(int)(ym-lxr>lup?ym-lxr:lup);
	int rl=(int)(ym+lxr<ldown?ym+lxr:ldown);
	lup=(unsigned int)ym;ldown=(unsigned int)ym;
	for(int i=(int)ym;i>=ll;i--)lup=dist[i]<dist[lup]?i:lup;
	for(i=(int)ym;i<=rl;i++)ldown=dist[i]<dist[ldown]?i:ldown;
	for(i=up;i<=(int)ldown;i++)sum+=YDIST[i];
	for(i=up;i<=(int)ldown;i++)retv+=(i-ym)*(i-ym)*YDIST[i]/sum;
	return sqrt(retv);
}

