// LaserPowerCalc.cpp: implementation of the CLaserPowerCalc class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "videoclnt.h"
#include "headers.h"
#include "MemoryManagement.h"
#include "LaserPowerCalc.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// CLaserPowerCalc - calculates the laser power according to a lookup-table
//                   with envelope-points kind of style

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// construction: initialize some variables
CLaserPowerCalc::CLaserPowerCalc()
{
 // set tables to none
 m_pSumPixels = NULL;
 m_pLaserPower = NULL;
 
 // not initialized
 m_bInitialized = FALSE;
}

// destruction: free up memory and deinitialize class
CLaserPowerCalc::~CLaserPowerCalc()
{
 // not initialized any more
 m_bInitialized = FALSE;
 
 // free up memory
 if (m_pSumPixels != NULL) memman.sw_free( (void *) m_pSumPixels );
 if (m_pLaserPower != NULL) memman.sw_free( (void *) m_pLaserPower );
}

//////////////////////////////////////////////////////////////////////
// Init() - initialize laser power lookup table to hardwired values
// 
// Return value:
//  TRUE  : successfully initialized
//  FALSE : error on init
//
BOOL CLaserPowerCalc::Init()
{
 if (m_bInitialized == TRUE) // already initialized ? 
 {
  // free up memory
  memman.sw_free( (void *) m_pSumPixels );
  m_pSumPixels = NULL;
  memman.sw_free( (void *) m_pLaserPower );
  m_pLaserPower = NULL;
 }

 // set entries to 11 (entries in predefined table)
 m_iEntries = 11;

 // allocate sum_pixels table
 m_pSumPixels = (unsigned int *) memman.sw_malloc( m_iEntries*sizeof(unsigned int) );
 if (m_pSumPixels == NULL) // error?
 {
  m_bInitialized = FALSE; 
  return FALSE;
 }
 
 // allocate laser power table
 m_pLaserPower = (double *) memman.sw_malloc( m_iEntries*sizeof(double) );
 if (m_pLaserPower == NULL) // error ? 
 {
  m_bInitialized = FALSE;
  return FALSE;
 }
 
 // set up predefined values
 m_pSumPixels[0]=0;
 m_pLaserPower[0]=0.0;

 m_pSumPixels[1]=1131909;
 m_pLaserPower[1]=70.0;

 m_pSumPixels[2]=1188226;
 m_pLaserPower[2]=40.0;

 m_pSumPixels[3]=2809968;
 m_pLaserPower[3]=1100.0;

 m_pSumPixels[4]=3062704;
 m_pLaserPower[4]=1200.0;

 m_pSumPixels[5]=3120117;
 m_pLaserPower[5]=1250.0;

 m_pSumPixels[6]=3130933;
 m_pLaserPower[6]=1050.0;

 m_pSumPixels[7]=3137667;
 m_pLaserPower[7]=1250.0;

 m_pSumPixels[8]=3192221;
 m_pLaserPower[8]=1250.0;

 m_pSumPixels[9]=4216262;
 m_pLaserPower[9]=1500.0;

 m_pSumPixels[10]=112412160;
 m_pLaserPower[10]=1500.0;

 // initialization was successful
 m_bInitialized = TRUE;

 return TRUE;

}

//////////////////////////////////////////////////////////////////////
// Init() - initialize laser power lookup table based on data in the 
//          file 'filename'
// 
// Return value:
//  TRUE  : successfully loaded and initialized
//  FALSE : error while loading
//
BOOL CLaserPowerCalc::Init( CString filename )
{
 // call LoadProfile method and return return value
 return( LoadProfile( filename ) );
}

//////////////////////////////////////////////////////////////////////
// LoadProfile() - load a laser power lookup table from file
// 
// Return value:
//  TRUE  : successfully loaded and initialized
//  FALSE : error while loading
//
BOOL CLaserPowerCalc::LoadProfile( CString filename )
{
 FILE *fp;
 char temp[200];
 char stop='\n';
 char *stopptr = &stop;

 // open file with "read text"
 fp = fopen( (LPCTSTR) filename, "rt");
 if (fp == NULL) return( FALSE ); // error? -> return false

 // read in first line
 fgets( temp, 199, fp);
 if (strncmp(temp, "LPLUT", 5) != 0)  // if first line is not starting with LPLUT -> return false
 {
  fclose( fp );
  return( FALSE );
 }

 // read in number of entries (data points) inside the file
 fgets( temp, 199, fp);
 int entries = atoi( temp );
 if (entries < 0 || entries > 1000000) // wrong number of entries -> return false
 {
  fclose( fp );
  return( FALSE );
 }
 
 // save entries into class variable
 m_iEntries = entries;

 // already initialized -> free up memory
 if (m_bInitialized == TRUE)
 {
  memman.sw_free( (void *) m_pSumPixels );
  m_pSumPixels = NULL;
  memman.sw_free( (void *) m_pLaserPower );
  m_pLaserPower = NULL;
 }
 
 // allocate sum_pixels array
 m_pSumPixels = (unsigned int *) memman.sw_malloc( m_iEntries*sizeof(unsigned int) );
 if (m_pSumPixels == NULL) // error ? 
 {
  m_bInitialized = FALSE;
  fclose( fp );
  return FALSE;
 }
 
 // allocate laser power array
 m_pLaserPower = (double *) memman.sw_malloc( m_iEntries*sizeof(double) );
 if (m_pLaserPower == NULL) // error ? 
 {
  m_bInitialized = FALSE;
  fclose(fp);
  return FALSE;
 }
 
 // read in number of entries
 int error = 0;
 for (int i=0;i<m_iEntries;i++)
 {
  char *err = fgets( temp, 199, fp);
  if (err == NULL) // error ?
  {
   error = 1;
   break;
  }
  // convert sum_pixels to unsigned long
  unsigned int sump = strtoul( temp, &stopptr, 10);
  err = fgets( temp, 199, fp);
  if (err == NULL) // error ? 
  {
   error = 1;
   break;
  }
  // convert to double
  double laserp = atof( temp );
  
  // store values
  m_pSumPixels[i] = sump;
  m_pLaserPower[i] = laserp;
 }

 // close file
 fclose( fp );
 if (error == 1) // error ? 
 {
  return( FALSE );
 }
 return( TRUE ); // success
}

//////////////////////////////////////////////////////////////////////
// CalcLaserPower() - calculates the laser power based on the 
//                    sum_pixels value
// 
// Return value:
//   laser power in microjoule
//
double CLaserPowerCalc::CalcLaserPower( unsigned int sum_pixels )
{
 // not yet initialized: return 0.0
 if (m_bInitialized == FALSE) return( 0.0 );

 // get the maximum number of pixels
 unsigned int maxpixels = m_pSumPixels[m_iEntries-1];
 // sum_pixels too big ? -> return 0.0
 if (sum_pixels > maxpixels) return( 0.0 );

 // seek through the table for the first sum_pixel entry bigger than sum_pixels
 for (int i=0;i<m_iEntries;i++)
 {
  if ( m_pSumPixels[i] > sum_pixels) break;
 }
 
 // obtain before and after sum of pixels
 unsigned int before = m_pSumPixels[i-1];
 unsigned int after  = m_pSumPixels[i];
 
 // obtain laser power on point 'before' and on point 'after'
 double beforeval = m_pLaserPower[i-1];
 double afterval = m_pLaserPower[i];

 // calculate m and n of linear function y=m*x+n

 double m = (afterval - beforeval) / (after*1.0-before*1.0);
 double n = (afterval) - (m*after*1.0);

 // calculate the laser power based on linear function
 double laserpower = m*sum_pixels + n;

 // return the laser power
 return( laserpower );
}
