// ServerList.cpp: implementation of the CServerList class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "videoclnt.h"
#include "Server.h"
#include "ServerList.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CServerList::CServerList()
{
    ::InitializeCriticalSection(&lock);    
	// make list empty (maybe not needed)
	m_ServerList.RemoveAll();
}


CServerList::~CServerList()
{
   // delete the servers
   deleteAllServers();
   // clean up synchronization stuff
   ::DeleteCriticalSection(&lock);
}


///////////////////////////////////////////////////////////////////////
// addServer() - adds a server to the end of the list of servers
//
// Return value:
//  TRUE: server was added
// FALSE: error, server was not added
//
BOOL CServerList::addServer( CServer srv )
{
 CServer *tmp;

 // lock access
 ::EnterCriticalSection(&lock);

 // create new server
 tmp = new CServer( srv.getPosition(), srv.getNickname(), srv.getServername(), srv.getListenport(), srv.getTineFrameEqpName(), 
                    srv.getCamString(), srv.getCamID(), srv.getTineControlName(), srv.getTineContext() );
 
 if (tmp == NULL) // error ? 
 {
  ::LeaveCriticalSection(&lock);
  return( FALSE ); // error
 }

 // add server to the end of the list
 m_ServerList.AddTail( (void *) tmp );

 ::LeaveCriticalSection(&lock);

 return( TRUE ); // success
}

///////////////////////////////////////////////////////////////////////
// deleteAllServers() - removes all servers from the list
//
// Return value:
//  TRUE: all servers were deleted (always true)
//
BOOL CServerList::deleteAllServers()
{
 CServer *tmp;
 
 //lock
 ::EnterCriticalSection(&lock);

 // loop through the list until it is empty
 while( m_ServerList.IsEmpty() == FALSE)
 {
  tmp = (CServer *) m_ServerList.RemoveHead(); //remove server
  delete tmp; // free memory allocated by server class
 }
 
 // removes all servers from the list (if there are some left)
 m_ServerList.RemoveAll();

 // unlock
 ::LeaveCriticalSection(&lock);
 
 //success
 return( TRUE );
}

///////////////////////////////////////////////////////////////////////
// getServer() - returns the server at 'position' inside the 'server'
//               reference
//
// Return value:
//  TRUE: server was returned
// FALSE: error, server was not returned (in most cases the server 
//               was simply not found)
//
BOOL CServerList::getServer( int position, CServer& server)
{
 BOOL retval=FALSE;

 // lock
 ::EnterCriticalSection(&lock);

 // empty list ? 
 if (m_ServerList.IsEmpty() == TRUE) 
 {
  ::LeaveCriticalSection(&lock);
  return FALSE; // error
 }

 // seek through the list
 for (int i=0;i<m_ServerList.GetCount();i++)
 {
  // get server at position i
  CServer *tmp = (CServer *) m_ServerList.GetAt( m_ServerList.FindIndex(i));
  if (tmp->getPosition() == position) // right server found ? 
  {
   // copy into 'server' reference parameter
   server.Reintialialize( tmp->getPosition(), tmp->getNickname(), tmp->getServername(), tmp->getListenport(), tmp->getTineFrameEqpName(), 
                    tmp->getCamString(), tmp->getCamID(), tmp->getTineControlName(), tmp->getTineContext());

   // success
   retval = TRUE;
  }
 }

 // unlock 
 ::LeaveCriticalSection(&lock);

 // return
 return( retval );
}

///////////////////////////////////////////////////////////////////////
// getServer() - returns the server with the given 'nickname' in the 'server'
//               reference
//
// Return value:
//  TRUE: server was returned
// FALSE: error, server was not returned (in most cases the server 
//               was simply not found)
//
BOOL CServerList::getServer( CString nickname, CServer& server)
{
 BOOL retval=FALSE;

 // lock 
 ::EnterCriticalSection(&lock);

 // list empty ? -> error
 if (m_ServerList.IsEmpty() == TRUE) 
 {
  //unlock 
  ::LeaveCriticalSection(&lock);
  return FALSE; // error
 }

 // seek through the list
 for (int i=0;i<m_ServerList.GetCount();i++)
 {
  // get server at specified position i
  CServer *tmp = (CServer *) m_ServerList.GetAt( m_ServerList.FindIndex(i));
  if (tmp->getNickname() == nickname) // right server found ? 
  {
   // copy into 'server' reference parameter
   server.Reintialialize( tmp->getPosition(), tmp->getNickname(), tmp->getServername(), tmp->getListenport(), tmp->getTineFrameEqpName(), 
                    tmp->getCamString(), tmp->getCamID(), tmp->getTineControlName(), tmp->getTineContext() );

   retval = TRUE;
  }
 }

 //unlock 
 ::LeaveCriticalSection(&lock);

 // return
 return( retval );

}

///////////////////////////////////////////////////////////////////////
// deleteServer() - deletes the server with the given 'serverpos' position
//
// Return value:
//  TRUE: server was deleted
// FALSE: error, server was not deleted (in most cases the server 
//               was simply not found)
// 
// Remarks:
//  Please do not forget to call Reinitialize after successfully 
//  deleting a server!
//
BOOL CServerList::deleteServer(int serverpos)
{
 BOOL retval=FALSE;

 // lock
 ::EnterCriticalSection(&lock);

 // server list empty ? -> error
 if (m_ServerList.IsEmpty() == TRUE) 
 {
  //unlock
  ::LeaveCriticalSection(&lock);
  return FALSE; //error
 }

 // seek through the list
 for (int i=0;i<m_ServerList.GetCount();i++)
 {
  // get the server
  CServer *tmp = (CServer *) m_ServerList.GetAt( m_ServerList.FindIndex(i));
  if (tmp->getPosition() == serverpos) // right server found?
  {
   // remove the server
   m_ServerList.RemoveAt( m_ServerList.FindIndex(i) );
   delete tmp; // free memory allocated by server class

   retval = TRUE;
  }
 }

 //unlock
 ::LeaveCriticalSection(&lock);

 //return
 return( retval );
}

///////////////////////////////////////////////////////////////////////
// Reorganize() - reorganizes the list after modifications
//
// Return value:
//  TRUE: returns always true
//
// Remarks:
//  Must be called after a successful call to deleteServer()
//
BOOL CServerList::Reorganize()
{
 BOOL retval=TRUE; // always true

 // lock 
 ::EnterCriticalSection(&lock);

 // empty list? -> we have nothing to do
 if (m_ServerList.IsEmpty() == TRUE) 
 {
  ::LeaveCriticalSection(&lock);
  return( retval ); // return
 }

 // seek through the list
 for (int i=0;i<m_ServerList.GetCount();i++)
 {
  // get the server
  CServer *tmp = (CServer *) m_ServerList.RemoveHead();

  CServer *pServer = new CServer();

  if (pServer == NULL) 
  {
	  ::LeaveCriticalSection(&lock);
	  return FALSE;
  }

  // copy server and change position number
  pServer->Reintialialize( i+1, tmp->getNickname(), tmp->getServername(), tmp->getListenport(), tmp->getTineFrameEqpName(), 
                    tmp->getCamString(), tmp->getCamID(), tmp->getTineControlName(), tmp->getTineContext() );

  // add to end of list
  m_ServerList.AddTail( (void *) pServer ); 

 }
 
 // unlock
 ::LeaveCriticalSection(&lock);


 // return
 return( retval );
}

int CServerList::Count() 
{
 int count = 0;
 // lock 
 ::EnterCriticalSection(&lock);

 // empty list? -> we have nothing to do
 if (m_ServerList.IsEmpty() != TRUE) 
 {
  count = m_ServerList.GetCount();
 }

 ::LeaveCriticalSection(&lock);

 return( count );
}

BOOL CServerList::modifyServer(int serverpos, CServer srv)
{
	BOOL retval=FALSE;

	// lock
	::EnterCriticalSection(&lock);

	// server list empty ? -> error
	if (m_ServerList.IsEmpty() == TRUE) 
	{
		//unlock
		::LeaveCriticalSection(&lock);
		return FALSE; //error
	}

	// seek through the list
	for (int i=0;i<m_ServerList.GetCount();i++)
	{
		// get the server
		CServer *tmp = (CServer *) m_ServerList.GetAt( m_ServerList.FindIndex(i));
		if (tmp->getPosition() == serverpos) // right server found?
		{
			 CServer *tmpNEW;

			 // create new server
			 tmpNEW = new CServer( tmp->getPosition(), srv.getNickname(), srv.getServername(), srv.getListenport(), srv.getTineFrameEqpName(), 
                    srv.getCamString(), srv.getCamID(), srv.getTineControlName(), srv.getTineContext() );
			 
			 // remove the server
			 m_ServerList.SetAt( m_ServerList.FindIndex(i), (void *) tmpNEW );
			 delete tmp; // free memory allocated by server class
		     retval = TRUE;
		}
	}

	//unlock
	::LeaveCriticalSection(&lock);

	//return
	return( retval );
}
