// Stat.cpp: implementation of the CStat class.
//
// Initially written by Velizar Miltchev (DESY Zeuthen) miltchev@ifh.de
// optimized, extended and some bugs fixed by Stefan Weisse sweisse@ifh.de
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "videoclnt.h"
#include "headers.h"
#include "Stat.h"
#include <assert.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////
// 
// The constructor of CStat does some basic initialization. In addition, the background image
// is subtracted, the area of interest is normalized and the image is (maybe) x-ray filtered.
// The projections are calculated, too. Please note that due to heavy optimization the ImWi
// parameter must be dividable by 8 without creating a fraction part.
//
// Parameters:
//  data     - pointer to the to be analysed image (must contain at least (ImWi*ImHi bytes)
//  back     - pointer to the to be subtracted background image (must contain at least ImWi*ImHi bytes)
//  out      - pointer where the analyzed image should be stored (must contain at least ImWi*ImHi bytes)
//  AOI      - rectangular region of area of interest inside frame - ranges must be x = [0,ImWi] y = [0,ImHi]
//  XFilt    - x-ray filtering TRUE=on FALSE=off
//  ImWi     - width of the image: _must_ be set to 768 (because of tight optimization)
//  ImHi     - height of the image: _must_ be set to 574 (because of tight optimization)
//
CStat::CStat(unsigned char* data,unsigned char* Back, unsigned char *out, CRect AOI,BOOL XFilt, BOOL circular, BOOL normalize, unsigned int ImWi,unsigned int ImHi, unsigned int bpp, unsigned int bpp_eff)
{
	// initialize member variables with the parameters
	filter=XFilt;
	array=data;
	BackGr=Back;
	ImageWidth=ImWi;
	ImageHeight=ImHi;
	m_iBpp = bpp;
	m_iBppEffective = bpp_eff;

    // initialize, normalize and clip area of interest
	if (circular == FALSE)
	{
	 AOI.NormalizeRect();
	 left=AOI.left;
	 right=AOI.right;
	 up=AOI.top;
	 down=AOI.bottom;
	 left=left*(left>0);
	 up=up*(up>0);
	 right=right*(right<(ImageWidth-1))+(ImageWidth-1)*(right>=(ImageWidth-1));
	 down=down*(down<ImageHeight-1)+(ImageHeight-1)*(down>=ImageHeight-1);
    }
	else
	{ // circular

	  // bounding rectangle of circle
      left = AOI.left - AOI.right;
	  up = AOI.top - AOI.right; 

      right = AOI.left + AOI.right;
	  down = AOI.top + AOI.right;  

      // fill proper circle values
	  center_x = AOI.left;
	  center_y = AOI.top;
	  length = AOI.right;

	  if (left < 0) left = 0;
	  if (left > ImageWidth-1) left = ImageWidth-1;
	  if (right < 0) right = 0;
	  if (right > ImageWidth-1) right = ImageWidth-1;
	  if (up < 0) up =0;
	  if (down < 0) down=0;
	  if (up > ImageHeight-1) up = ImageHeight-1;
	  if (down > ImageHeight-1) down = ImageHeight-1;

	  if (length < 3) length = 3;
	}

    // initialize to-be-calculated member variables
	XMEAN=0;
	YMEAN=0;
	XRMS=0;
	YRMS=0;

    // allocate arrays needed for distribution calculation
	YDIST=   (double *) memman.sw_malloc( sizeof(double)*ImageHeight );
	XDIST=   (double *) memman.sw_malloc( sizeof(double)*ImageWidth );
    YDISTd = (double *) memman.sw_malloc( sizeof(double)*ImageHeight );
    XDISTd = (double *) memman.sw_malloc( sizeof(double)*ImageWidth );

	if (XDIST == NULL || YDIST == NULL || XDISTd == NULL || YDISTd == NULL)
	{
    	return;
	}

    for (int i=0;i<ImageWidth;i++)
	{
     XDIST[i] = 0.0;
	 XDISTd[i] = 0.0;
	}

    for (i=0;i<ImageHeight;i++)
	{
     YDIST[i] = 0.0;
	 YDISTd[i] = 0.0;
	}

    memman.memcpysw( (char *) out, (char *) array, ImageWidth*ImageHeight*m_iBpp/8 );

	if (filter == TRUE) XFilter( out );

	BackSubtraction( out , BackGr );

	CalcProjection( out, circular );

	//DownScaleTo8Bit( temparr, out, ImageWidth, ImageHeight, m_iBpp );

	if (normalize == TRUE) NormalizeImage( out, circular );

}


///////////////////////////////////////////////////////////////////////////////////////////
// 
// In the destructor just the dynamically allocated variables are freed.
//
CStat::~CStat()
{
    // delete x projection arrays
	if (XDIST != NULL) memman.sw_free( (void *) XDIST );
	if (XDISTd != NULL) memman.sw_free( (void *) XDISTd );

    // delete y projection arrays
	if (YDIST != NULL) memman.sw_free( (void *) YDIST );
	if (YDISTd != NULL) memman.sw_free( (void *) YDISTd );
}

///////////////////////////////////////////////////////////////////////////////////////////
// 
// In this function, the spot center coordinates and size of the spot is calculated. After 
// calling this function, one can get the calculated value by calling functions like X_RMS(),
// Y_RMS, XM() and YM().
//
void CStat::START()
{
	if (XDIST == NULL || YDIST == NULL) return;
	XMEAN=CalcXMEAN(XDIST);
	YMEAN=CalcYMEAN(YDIST);
	XRMS=CalcXRMS(XDIST,XMEAN);
	YRMS=CalcYRMS(YDIST,YMEAN);

}

////////////////////////////////////////////////////////////////////////
// DownScaleTo8Bit() 
//
// This mathod scales image src (attributes: width, height and bpp) into 
// 8 bits memory location dst (attributes: width, height, bpp=8).
//
// One has to take care that memory location src is at least as big as 
//  width*height*bpp/8 bytes
// and that dst is at least as big as
//  width*height bytes
//
//DEL void CStat::DownScaleTo8Bit(unsigned char *src, unsigned char *dst, unsigned int width, unsigned int height, unsigned int bpp)
//DEL {
//DEL 
//DEL  if (bpp == 8)  // 8 bits per pixel, then just copy the data
//DEL  {
//DEL   memman.memcpysw( (char *) dst, (char *) src, width*height);
//DEL  }
//DEL  else if (bpp == 16) // 16 bits per pixel 
//DEL  {
//DEL   unsigned short maxval=0;
//DEL   unsigned short *srci = (unsigned short *) src;
//DEL 
//DEL   // look into source for maximum value
//DEL   for (unsigned int j=0;j<width*height;j++)
//DEL   {
//DEL    if (srci[j] > maxval) maxval = srci[j];
//DEL   }
//DEL   
//DEL   // scale into 8 bits accordingly to the maximum value
//DEL   unsigned long tmp = 65536L / (((unsigned long)maxval)+1);
//DEL   unsigned short tmp1 = (unsigned short) tmp;
//DEL   for (j=0;j<width*height;j++)
//DEL   {
//DEL    dst[j] = (unsigned char) ((srci[j] * tmp1) >> 8);
//DEL   }
//DEL  }
//DEL  else if (bpp == 32) // 32 bits per pixel
//DEL  {
//DEL   unsigned long *srcl = (unsigned long *) src;
//DEL   unsigned long maxval=0;
//DEL 
//DEL   // look into source for maximum value
//DEL   for (unsigned int j=0;j<width*height;j++)
//DEL   {
//DEL    if (srcl[j] > maxval) maxval = srcl[j];
//DEL   }
//DEL   
//DEL   // scale into 8 bits accordingly to the maximum value
//DEL   unsigned long tmp = 0xFFFFFFFF / (maxval);
//DEL   for (j=0;j<width*height;j++)
//DEL   {
//DEL    dst[j] = (unsigned char) ((srcl[j] * tmp) >> 24);
//DEL   }
//DEL  
//DEL  }
//DEL 
//DEL }


///////////////////////////////////////////////////////////////////////////////////////////
//
// This function does image processing. Firstly, the background image is subtracted from the
// video frame. After this, the maximum pixel value in the area of interest is determined.
// The maximum value is then used to normalize the area of interest, e.g. if the brightest 
// pixel has a value of 127 all pixels which have the same value will be scaled to 255. 
// Smaller values are scaled according to the same factor. In the same inner loop we are 
// normalizing the frame we are also calculating XDISTi and YDISTi integer arrays. Because
// we walk over the data three times (but depends on how big the area of interest is), this 
// function costs a lot of cpu cycles.
//
// outdata has same bits-width as datablock, need to downscale to 8bit if necessary
void CStat::NormalizeImageRectangular(void *datablock)
{
    unsigned int effective_colorvalues = (1 << m_iBppEffective);

	if (m_iBpp == 8)
	{
		static unsigned char table[256];
		int xlen = right - left;
		int ylen = down - up;
		register unsigned int xadd = (ImageWidth-1) - (right - left);
		register unsigned int xstart = up*ImageWidth+left;
		int xpixels = right - left + 1;
		int ypixels = down - up + 1;
 
		if (xlen < 1 || ylen < 1) return;

		register unsigned char *data = (unsigned char *) datablock;
		unsigned char maxnum = 0;

		for (register int y=0;y<ypixels;y++)
		{
			for (register int x=0;x<xpixels;x++)
			{
				if (data[xstart] > maxnum) maxnum = data[xstart];
				xstart++;
			}
			xstart+=xadd;
		}
    

		// build table for faster processing in the next inner loop 

		if (maxnum == 0) maxnum = 1;
		double val = ((double)(effective_colorvalues-1)) / (1.0*maxnum);

		for (int m=0;m<256;m++)
		{
			table[m] = (unsigned char) (((double)m)*val);
		}
    
		// normalize the area of interest
		// and calculata XDIST and YDIST (integer mode)

		xstart = up*ImageWidth+left;
    
		// with normalization
		for (y=up;y<=down;y++)
		{
			for (register int x=left;x<=right;x++)
			{
				// normalisation
				data[xstart] = table[data[xstart]];
				xstart++;
			}
			xstart += xadd;
		}
	}
	else if (m_iBpp == 16) // will be normalize to effective bpp later !!! SW! Sept 20, 2006
	{
		static unsigned short table[256*256];
		int xlen = right - left;
		int ylen = down - up;
		register unsigned int xadd = (ImageWidth-1) - (right - left);
		register unsigned int xstart = up*ImageWidth+left;
		int xpixels = right - left + 1;
		int ypixels = down - up + 1;
 
		if (xlen < 1 || ylen < 1) return;

		register unsigned short *data = (unsigned short *) datablock;
		unsigned short maxnum = 0;

		for (register int y=0;y<ypixels;y++)
		{
			for (register int x=0;x<xpixels;x++)
			{
				if (data[xstart] > maxnum) maxnum = data[xstart];
				xstart++;
			}
			xstart+=xadd;
		}
    

		// build table for faster processing in the next inner loop 

		if (maxnum == 0) maxnum = 1;
		double val = ((double)(effective_colorvalues-1)) / (1.0*maxnum);

		for (int m=0;m<65536;m++)
		{
			table[m] = (unsigned short) (((double)m)*val);
		}
    
		// normalize the area of interest
		// and calculata XDIST and YDIST (integer mode)

		xstart = up*ImageWidth+left;
    
		// with normalization
		for (y=up;y<=down;y++)
		{
			for (register int x=left;x<=right;x++)
			{
				// normalisation
				data[xstart] = table[data[xstart]];
				xstart++;
			}
			xstart += xadd;
		}

	}
}

void CStat::NormalizeImageCircular(void *datablock)
{
	unsigned int effective_colorvalues = (1 << m_iBppEffective);

	if (m_iBpp == 8)
	{
		static unsigned char table[256];
		int xlen = right - left;
		int ylen = down - up;
		register unsigned int xadd = (ImageWidth-1) - (right - left);
		register unsigned int xstart = up*ImageWidth+left;
		unsigned int xpixels = right - left + 1;
		unsigned int ypixels = down - up + 1;
 
		if (xlen < 1 || ylen < 1) return;

		register unsigned char *data = (unsigned char *) datablock;
		unsigned char maxnum = 0;

		for (register int y=up;y<=down;y++)
		{
			int locleft; 
			int locright; 

			GetLeftAndRight(y, locleft, locright);

			if ((locleft == 0) && (locright == 0))
			{
    
			}
			else
			{
				xstart = y*ImageWidth+locleft;
				for (register int x=locleft;x<=locright;x++)
				{
					if (data[xstart] > maxnum) maxnum = data[xstart];
					xstart++;
				}
			}
		}

		// build table for faster processing in the next inner loop 
		if (maxnum == 0) maxnum = 1;
		double val = ((double)(effective_colorvalues-1)) / (1.0*maxnum);

		for (int m=0;m<256;m++)
		{
			table[m] = (unsigned char) (((double)m)*val);
		}

		// normalize the area of interest
		// and calculata XDIST and YDIST (predouble mode)

		xstart = up*ImageWidth+left;
 
		// with normalization
		for (y=up;y<=down;y++)
		{
			int locleft;
			int locright;

			GetLeftAndRight(y, locleft, locright);

			if (locleft < 0) assert(1==2);
			if (locright > ((int)ImageWidth-1)) assert(1==9);

			if ((locleft==0) && (locright==0))
			{
     
			}
			else
			{
				xstart=y*ImageWidth+locleft; 
				for (register int x=locleft;x<=locright;x++)
				{
					// normalization
					data[xstart] = table[data[xstart]];
 
					xstart++;
				}
			}
		}
	}
	else if (m_iBpp == 16)
	{
		static unsigned short table[256*256];
		int xlen = right - left;
		int ylen = down - up;
		register unsigned int xadd = (ImageWidth-1) - (right - left);
		register unsigned int xstart = up*ImageWidth+left;
		unsigned int xpixels = right - left + 1;
		unsigned int ypixels = down - up + 1;
 
		if (xlen < 1 || ylen < 1) return;

		register unsigned short *data = (unsigned short *) datablock;
		unsigned short maxnum = 0;

		for (register int y=up;y<=down;y++)
		{
			int locleft; 
			int locright; 

			GetLeftAndRight(y, locleft, locright);

			if ((locleft == 0) && (locright == 0))
			{
    
			}
			else
			{
				xstart = y*ImageWidth+locleft;
				for (register int x=locleft;x<=locright;x++)
				{
					if (data[xstart] > maxnum) maxnum = data[xstart];
					xstart++;
				}
			}
		}

		// build table for faster processing in the next inner loop 
		if (maxnum == 0) maxnum = 1;
		double val = ((double)(effective_colorvalues-1)) / (1.0*maxnum);

		for (int m=0;m<65536;m++)
		{
			table[m] = (unsigned char) (((double)m)*val);
		}

		// normalize the area of interest
		
		xstart = up*ImageWidth+left;
 
		// with normalization
		for (y=up;y<=down;y++)
		{
			int locleft;
			int locright;

			GetLeftAndRight(y, locleft, locright);

			if (locleft < 0) assert(1==2);
			if (locright > ((int)ImageWidth-1)) assert(1==9);

			if ((locleft==0) && (locright==0))
			{
     
			}
			else
			{
				xstart=y*ImageWidth+locleft; 
				for (register int x=locleft;x<=locright;x++)
				{
					// normalization
					data[xstart] = table[data[xstart]];
 
					xstart++;
				}
			}
		}
	}
}


void CStat::NormalizeImage( void *inout, BOOL circ )
{
 if (circ == TRUE)
 {
	 NormalizeImageCircular( inout);
 }
 else
 {
     NormalizeImageRectangular( inout);
 }
}



////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot center on the x axis out of the 
// distribution array that is passed as 'dist'.
//
double  CStat::CalcXMEAN(double* dist)
{
	double RV=0;
	for(int i=left;i<=right;i++)
		RV+=(double)i*dist[i];
	return RV;
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot center on the Y axis out of the 
// distribution array that is passed as 'dist'.
//
double  CStat::CalcYMEAN(double* dist)
{
	double RV=0;
	for(int i=up;i<=down;i++)
		RV+=(double)i*dist[i];
	return RV;
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot size on the X axis out of the 
// distribution array that is passed as 'dist' and the x mean value passed
// as 'xm'.
//
double  CStat::CalcXRMS(double* dist,double xm)
{
	double RV=0;
	int l=right-left;
	for(int i=left;i<=right;i++)RV+=(i-xm)*(i-xm)*dist[i];
	//FitAndSpline(&XDIST[left],&XDIST[left],l,1+9*(l>10));
	return sqrt(RV);
}

////////////////////////////////////////////////////////////////////////////
//
// This function calculates the spot size on the Y axis out of the 
// distribution array that is passed as 'dist' and the y mean value passed
// as 'ym'.
//
double  CStat::CalcYRMS(double* dist,double ym)
{
	double RV=0;
	int l=down-up;
	for(int i=up;i<=down;i++)RV+=(i-ym)*(i-ym)*dist[i];
	//FitAndSpline(YDIST,YDIST,l,1+9*(l>10));
	return sqrt(RV);
}

////////////////////////////////////////////////////////////////////////////
//
// This function does x-ray filtering of the image stored in 'data'. It is 
// working like a 'discard-single-peak' algorithm. The idea of this algorithm 
// is to discard pixels of non-zero value that are surrounded by zero value 
// (black colored) pixels. In addition one can generalize this kind of logic 
// and make all pixels, which stick up more than p=20% over their neighbors, 
// equal to the average of the surrounding pixels.
//

void CStat::XFilter(void *data)
{
  if (m_iBpp == 8)
  {
    unsigned char *data8 = (unsigned char *)data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		int equalize=0;
		const int datapos = data8[pos];
		if(datapos)
		{
			const int datapos2 = (datapos * 2) / 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=((data8[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data8[pos+ImageWidth+1])<datapos2);


			if(equalize>3)
  /*equalize*/{
				unsigned int sum=0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data8[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum >>= 3;
				
				data8[pos]=(unsigned char)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }
  else if (m_iBpp == 16)
  {
    unsigned short *data16 = (unsigned short *) data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		int equalize=0;
		const int datapos = data16[pos];
		if(datapos)
		{
			const unsigned int datapos2 = (datapos * 2) / 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=((data16[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=((data16[pos+ImageWidth+1])<datapos2);


			if(equalize>3)
  /*equalize*/{
				unsigned int sum=0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=data16[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum >>= 3;
				
				data16[pos]=(unsigned short)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }
  else // bpp == 32
  {
    unsigned int *data32 = (unsigned int *) data;
	for(int j=up+1;j<=down-1;j++)
	{
	 int pos = j*ImageWidth+left+1;
	 
	 for( int i=left+1;i<=right-1;i++)
	 {
		double equalize=0;
		const unsigned int datapos = data32[pos];
		if(datapos)
		{
			const unsigned int datapos2 = (datapos / 2) * 3;
			//lr = -1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
              equalize+=(double) ((data32[pos-ImageWidth-1])<datapos2);

			//lr = -1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-1])<datapos2);

			//lr = -1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth-1])<datapos2);

			//lr=0;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-ImageWidth])<datapos2);

			//lr = 0;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth])<datapos2);

			//lr = 1;
			//ub = -1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos-ImageWidth+1])<datapos2);

			//lr = 1;
			//ub = 0;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+1])<datapos2);

			//lr = 1;
			//ub = 1;
			//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
			equalize+=(double) ((data32[pos+ImageWidth+1])<datapos2);


			if(equalize>3.0)
  /*equalize*/{
				double sum=0.0;

  			    //lr = -1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth-1];

			    //lr = -1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-1];

			    //lr = -1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth-1];

			    //lr=0;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth];

			    //lr = 0;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth];

			    //lr = 1;
			    //ub = -1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos-ImageWidth+1];

			    //lr = 1;
			    //ub = 0;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+1];

			    //lr = 1;
			    //ub = 1;
			    //sum+=data[(i+lr)+(j+ub)*ImageWidth];
                  sum+=(double) data32[pos+ImageWidth+1];
				
				//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
				sum /= 8.0;
				
				data32[pos]=(unsigned int)sum;
  /*equalize*/}
          }
		pos++;
			
	 }
	 
	}
  }

}

/*
void CStat::XFilter(unsigned char *data)
{ 
	for(unsigned int i=left+1;i<=right-1;i++)
	 for(unsigned int j=up+1;j<=down-1;j++)
	 {
		int discard=0;
		if(data[i+ImageWidth*j])
			for(int lr=-1;lr<=1;lr++)
			for(int ub=-1;ub<=1;ub++)
			{
				if(lr*lr+ub*ub)
				{
					discard+=(data[(i+lr)+(j+ub)*ImageWidth]==0);
				}
			}
			if(discard>6)data[i+ImageWidth*j]=0;
	 }
	for( i=left+1;i<=right-1;i++)
	 for(unsigned int j=up+1;j<=down-1;j++)
	 {
		int equalize=0;
		if(data[i+ImageWidth*j])
			for(int lr=-1;lr<=1;lr++)
			for(int ub=-1;ub<=1;ub++)
			{
				if(lr*lr+ub*ub)
				{
					equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
				}
			}
			if(equalize>6)
//equalize  {
				int sum=0;
				for(int lr=-1;lr<=1;lr++)
				for(int ub=-1;ub<=1;ub++)
				{
					if(lr*lr+ub*ub)
					{
						sum+=data[(i+lr)+(j+ub)*ImageWidth];
					}
				}
				sum/=8;
				data[i+ImageWidth*j]=(unsigned char)sum;
//equalize  }
			
	 }
	 for(i=left;i<=right;i++)
	 {
		 data[i]=0;
		 data[i+ImageWidth*(ImageHeight-1)]=0;
	 }
	 for(unsigned int j=up;j<=down;j++)
	 {
		 data[j*ImageWidth]=0;
		 data[ImageWidth-1+ImageWidth*(ImageHeight-1)]=0;
	 }

}
*/

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the XDIST distribution array. Plase note that you 
// obtain a pointer to an internal array. So after destruction of this class the 
// array is no longer valid !
//
double*  CStat::X_DIST()
{
	return XDIST;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the YDIST distribution array. Plase note that you 
// obtain a pointer to an internal array. So after destruction of this class the 
// array is no longer valid !
//
double*  CStat::Y_DIST()
{
	return YDIST;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot size in x dimension. You should only call
// this function after calling START().
//
double  CStat::X_RMS()
{
	return XRMS;
}

///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot size in y dimension. You should only call
// this function after calling START().
//
double  CStat::Y_RMS()
{
	return YRMS;
}


///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot center in x dimension. You should only call
// this function after calling START().
//
double  CStat::XM()
{
	return XMEAN;
}


///////////////////////////////////////////////////////////////////////////////////
//
// This public function returns the spot center in y dimension. You should only call
// this function after calling START().
//
double  CStat::YM()
{
	return YMEAN;
}



unsigned int CStat::SUM_PIXELS()
{
 return( m_uiSumPixels );
}

// uses center_x, center_y and length
void CStat::GetLeftAndRight( int ypos, int &leftval, int &rightval)
{
 if ((center_y-length) > ypos) 
 {
  leftval =0;
  rightval = 0;
  return;
 }
 if ((center_y+length) < ypos) 
 {
  leftval = 0;
  rightval = 0;
  return;
 }

 double value = sqrt((1.0*length*length) - (1.0*(ypos-center_y)*(ypos-center_y)));

 leftval = center_x - ((int) value);
 rightval = center_x + ((int) value);
 if (leftval < 0) leftval = 0;
 if (rightval >= ImageWidth) rightval = ImageWidth-1;
}

void CStat::BackSubtraction(unsigned char *inout_buffer, unsigned char *backbuf)
{
 unsigned char *datablock = (unsigned char *) inout_buffer;
 unsigned char *background = backbuf;
 unsigned char *outptr = inout_buffer;

 unsigned int pixelsdiv8 = ((ImageWidth*ImageHeight*m_iBpp)/8) / 8;

 BOOL mmxdetected=FALSE;

 // subtract background from whole image, to speed up the process we use MMX
 // instructions and process 8 pixels at once
 // if you can not use MMX for some reason or need to be portable use this 
 // code snippet (ANSI C++):
 //

/*
 unsigned char *dataptr = datablock;
 unsigned char *backgnd = (unsigned char *) BackGr;
 unsigned char *outptr = outdata;
 
 for (unsigned int i=0;i<ImageHeight*ImageWidth;i++)
 {
  int pix = (int) dataptr[i];
  int back = (int) backgnd[i];
  int out;

  if ((out = pix-back) < 0) out = 0;
  
  outptr[i] = (unsigned char) out;
 }
*/

 __asm
 {
     push ebx
	 mov EAX, 1 
     CPUID 
	 pop ebx
     test EDX, 00800000h 
 
     jz MMX_Technology_Not_Found
     mov mmxdetected, 1
     MMX_Technology_Not_Found:
 }

 if (mmxdetected == TRUE)
 {

   __asm
   {
	 emms

     pxor MM1, MM1
	 movq MM2, MM1
	 movq MM3, MM1

	 mov esi, datablock
	 mov edi, outptr
	 mov ebx, background

	 mov ecx, pixelsdiv8 // 768*574 / 8
    ploop1: 

 	 movq MM2, [ebx]
	 movq MM1, [esi] 
	 add  ebx,8

	 psubusb MM1, MM2
	 
	 movq [edi], MM1
	 add  esi,8
	 add  edi,8

	 dec ecx
  	 jnz ploop1

	 emms
   }
 }
  else
 {
   if (m_iBpp == 8)
   {
	  const unsigned char *dataptr = inout_buffer;
      const unsigned char *backgnd = (unsigned char *) backbuf;
      unsigned char *outptr = (unsigned char *) inout_buffer;
 
      for (int i=0;i<ImageHeight*ImageWidth;i++)
	  {
        int pix = (int) dataptr[i];
        int back = (int) backgnd[i];
        int out;

        if ((out = pix-back) < 0) out = 0;
  
        outptr[i] = (unsigned char) out;
	  }
   }
   else if (m_iBpp == 16)
   {
	  const unsigned short *dataptr = (unsigned short *) inout_buffer;
      const unsigned short *backgnd = (unsigned short *) backbuf;
      unsigned short *outptr = (unsigned short *) inout_buffer;
 
      for (int i=0;i<ImageHeight*ImageWidth;i++)
	  {
        int pix = (int) dataptr[i];
        int back = (int) backgnd[i];
        int out;

        if ((out = pix-back) < 0) out = 0;
  
        outptr[i] = (unsigned short) out;
	  }
   }
   else // bpp == 32
   {
	  const unsigned long *dataptr = (unsigned long *) inout_buffer;
      const unsigned long *backgnd = (unsigned long *) backbuf;
      unsigned long *outptr = (unsigned long *) inout_buffer;
 
      for (int i=0;i<ImageHeight*ImageWidth;i++)
	  {
        double pix = (double) dataptr[i];
        double back = (double) backgnd[i];
        double out;

        if ((out = pix-back) < 0) out = 0.0;
  
        outptr[i] = (unsigned long) out;
	  }
   }

 }

}

//////////////////////////////////////////////////////////////////////////////////////////
// OUTDATED DOC!!!
// When this function is called, we already have created the XDISTi und YDISTi arrays.
// Now we can calculate the real XDIST and YDIST arrays out of it.
//
// Please note that the calculation is a bit tricky, to let it work even when there is a 
// totally black picture. Because of this, we are adding some static noise there (notice 
// the 1's and 1.0's that should not be there) to not falsify the calculations.
//
// Background:
//
//  Sometimes it happens that you don't have data on all pictures, but only on some of them.
//  The other pictures are totally black. You must know that on totally black pictures,
//  the mean and rms sizes are not calculated right. The spot is always in the top left corner
//  of the video window if the frame is black. So if you grab 20 pictures and average the 
//  data analysis over all pictures, the spot would be located on the totally wrong position.
//  Now it is at least in the middle of the area of interest.
//
void CStat::CalcProjection(unsigned char *in, BOOL circular)
{
 // circular not used up to now

 assert( right < ImageWidth);
 assert( left >= 0);
 assert( left < ImageWidth );
 assert( right >= 0);
 assert( up < ImageHeight );
 assert( up >= 0);
 assert( down < ImageHeight);
 assert( down >= 0);

 register unsigned int xadd = (ImageWidth-1) - (right - left);
 register unsigned int xstart = up*ImageWidth+left;
 register double sum;

 if (m_iBpp == 8)
 {
   unsigned char *data = in;
	 // with normalization
   for (register int y=up;y<=down;y++)
   {
    sum = 0.0;
    for (register int x=left;x<=right;x++)
    {
     register unsigned int p;
     p = (unsigned int) data[xstart];
     // calculation of nDISTi
     sum += (double) p;
     XDISTd[x] += (double) p;
 
     xstart++;
    }
    xstart += xadd;
    YDISTd[y] = sum;
   }
 }
 else if (m_iBpp == 16)
 {
   unsigned short *data = (unsigned short *) in;
   // with normalization
   for (register int y=up;y<=down;y++)
   {
    sum = 0.0;
    for (register int x=left;x<=right;x++)
    {
     register unsigned int p;
     p = (unsigned int) data[xstart];
     // calculation of nDISTi
     sum += (double) p;
     XDISTd[x] += (double) p;
 
     xstart++;
    }
    xstart += xadd;
    YDISTd[y] = sum;
   }
 }
 else // bpp == 32
 {
   unsigned long *data = (unsigned long *) in;
   // with normalization
   for (register int y=up;y<=down;y++)
   {
    sum = 0.0;
    for (register int x=left;x<=right;x++)
    {
     register unsigned int p;
     p = (unsigned int) data[xstart];
     // calculation of nDISTi
     sum += (double) p;
     XDISTd[x] += (double) p;
 
     xstart++;
    }
    xstart += xadd;
    YDISTd[y] = sum;
   }
 }

 int sumi=0;

 sum = 1.0;
 for (register int x=(int)left;x<=(int)right;x++)
 {
  sum += XDISTd[x];
 }

 m_uiSumPixels = (unsigned int) sum;

 if (sum > 0.9999 && sum < 1.0001) { sumi = 1; } else { sum-=1.0; }

 double ZEROh = 1.0 * (double)(sumi==1)/ ((double) (right-left+1));
 double ZEROv = 1.0 * (double)(sumi==1)/ ((double) (down-up+1));

 // scale xdist accordingly
 for (x=(int)left;x<=(int)right;x++)
 {
  XDIST[x] = (ZEROh+(double)XDISTd[x])/((double) sum);
 }

 // scale ydist accordingly
 for (int y=up;y<=down;y++)
 {
  YDIST[y] = (ZEROv+(double)YDISTd[y])/((double) sum);
 }
}
